from datetime import datetime
import numpy as np

from iss_lidar.tools import create_filename, wspd_wdir_from_uv


def test_create_filename():
    # VAD
    destdir = "/scr/rain-isf/apg/projects/swex_2022/iss1/ds/lidar/vad"
    fname = destdir + "/VAD_339_20220329.nc"
    dt = datetime(2022, 3, 29)
    assert create_filename(dt, destdir, "VAD", "339") == fname
    # without prefix
    fname = destdir + "/VAD_20220329.nc"
    assert create_filename(dt, destdir, "VAD", None) == fname
    # destdir with trailing /
    destdir = "/scr/rain-isf/apg/projects/swex_2022/iss1/ds/lidar/vad/"
    assert create_filename(dt, destdir, "VAD", None) == fname
    # consensus
    destdir = "/scr/rain-isf/apg/projects/swex_2022/iss1/ds/lidar/consensus"
    fname = destdir + "/30min_winds_339_20220329.nc"
    assert create_filename(dt, destdir, "30min_winds", "339") == fname
    # without prefix
    fname = destdir + "/30min_winds_20220329.nc"
    assert create_filename(dt, destdir, "30min_winds", None) == fname


def test_wind_from_uv():
    u = np.array([-1, np.NZERO, 0, 1, 0, -2, np.nan], dtype=float)
    v = np.array([np.NINF, np.NZERO, 0, 0, -1, -2, 2], dtype=float)
    xspd = np.array([np.inf, 0, 0, 1, 1, np.sqrt(8), np.nan], dtype=float)
    xdir = np.array([0, 90, 270, 270, 0, 45, np.nan], dtype=float)
    wspd, wdir = wspd_wdir_from_uv(u, v)
    np.testing.assert_allclose(wspd, xspd, equal_nan=True)
    np.testing.assert_allclose(wdir, xdir, equal_nan=True)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# pylint: disable=C0103,E1101

"""
Decoder for raw .hpl files that are outputted by the Halo Photonics doppler
lidar. Many things are hard coded and should be documented with in-line
comments. Outputs decoded data to netcdf-4. May have to change the 'lookup'
variable to account for different header info. Though not originally coded to
decode RHI scans, it shouldn't be a problem as long as it is specified in the
lookup table


Original Author: Tyler Bell - University of Oklahoma (March 2017)
"""


import logging
import os
import argparse
from datetime import datetime, timedelta
import netCDF4
import numpy as np


# Key is name of scan type in header, value is name as it should appear
# in the filename
lookup = {'User file 4 - stepped': 'fp',
          'User file 3 - stepped': 'ppi',
          'User file 2 - stepped': 'fp',
          'User file 1 - stepped': 'ppi',
          'csm_Tail_7VT_Coplanar - csm': 'Lionstail',
          'csm_George - csm': 'George',
          'VAD - stepped': 'ppi',
          'Stare - stepped': 'fp',
          'Stare': 'fp'}

FILL_VALUE = -9999


def decode_header(header):
    """
    Takes in a list of lines from the raw hpl file. Separates them by
    tab and removes unnecessary text
    """
    new_header = {}

    for item in header:
        split = item.split('\t')
        # new_header[split[0].replace(':', '')] = split[1].replace("\r\n", "")
        new_header[split[0].replace(':', '')] = split[1].strip()

    return new_header


def _to_epoch(dt):
    return (dt - datetime(1970, 1, 1)).total_seconds()


def process_file(in_file, out_dir, prefix):
    """
    Processes a raw halo hpl file and turns it into a netcdf
    :param in_file:
    :param out_dir:
    :return:
    """

    # Read in the text file
    lines = []
    with open(in_file) as f:
        for line in f:
            lines.append(line)

    logging.debug("Decoding header")
    # Read in the header info
    header = decode_header(lines[0:11])

    ngates = int(header['Number of gates'])
    # nrays = int(header['No. of rays in file'])
    # Cant do this apparently. Not always correct (wtf)
    len_data = len(lines[17:])
    nrays = int(len_data / (ngates + 1))

    gate_length = float(header['Range gate length (m)'])
    start_time = datetime.strptime(header['Start time'], '%Y%m%d %H:%M:%S.%f')
    scan_type = lookup[header['Scan type']]

    logging.info("Processing file type: %s" % scan_type)
    logging.debug("Number of rays: %s" % nrays)

    logging.debug("Reading data")
    # Read in the actual data
    az = np.zeros(nrays)
    hour = np.zeros(nrays)
    elev = np.zeros(nrays)
    pitch = np.zeros(nrays)
    roll = np.zeros(nrays)
    rng = np.asarray([(gate + .5) * gate_length for gate in range(ngates)])

    vel = np.zeros((ngates, nrays))
    intensity = np.zeros((ngates, nrays))
    beta = np.zeros((ngates, nrays))

    try:
        for ray in range(nrays):
            # Get the scan info
            info = lines[ray * (ngates + 1) + 17].split()
            hour[ray] = float(info[0])
            az[ray] = float(info[1])
            elev[ray] = float(info[2])
            pitch[ray] = float(info[3])
            roll[ray] = float(info[4])

            for gate in range(ngates):
                data = lines[ray * (ngates + 1) + 17 + gate + 1].split()
                vel[gate, ray] = float(data[1])
                intensity[gate, ray] = float(data[2])
                beta[gate, ray] = float(data[3])

    except IndexError:
        logging.warning("Something went wrong with the indexing here...")

    logging.debug('Preparing to write netcdf')

    # Get the times and dates figured out for the netcdf
    time = []
    epoch = []
    for h in hour:
        dt = datetime(start_time.year, start_time.month,
                      start_time.day) + timedelta(hours=h)
        time.append(dt)
        epoch.append(_to_epoch(dt))

    time = np.asarray(time)
    epoch = np.asarray(epoch)
    base_time = _to_epoch(start_time)
    time_offset = epoch - base_time

    # Figure out netcdf attrs
    nc_attrs = {'start_time': start_time.strftime('%Y-%m-%dT%H:%M:%S')}
    # None right now

    # Get the filename figured out
    if prefix is None:
        filename = start_time.strftime(
            "{type}_%Y%m%d_%H%M%S.nc".format(type=scan_type))
    else:
        filename = start_time.strftime(
            "{prefix}_{type}_%Y%m%d_%H%M%S.nc".format(type=scan_type,
                                                      prefix=prefix))

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    filename = os.path.join(out_dir, filename)

    # Write out the netcdf
    logging.info("Writing netcdf")
    logging.debug("Creating file: {}".format(filename))

    nc = netCDF4.Dataset(filename, "w", format="NETCDF4")

    # Create the dimensions
    nc.createDimension('time', size=None)
    nc.createDimension('range', size=len(rng))

    # Set the netcdf attributes
    logging.debug('Writing attributes')
    nc.setncatts(nc_attrs)

    logging.debug('Writing base_time')
    var = nc.createVariable('base_time', 'i8')
    var.setncattr('long_name', 'Time')
    var.setncattr('units', 'seconds since 1970-01-01 00:00:00 UTC')
    var[:] = base_time

    logging.debug('Writing time_offset')
    var = nc.createVariable('time_offset', 'i8', dimensions=('time',))
    var.setncattr('long_name', 'Time offset')
    var.setncattr('unis', 'seconds since base_time')
    var[:] = time_offset

    logging.debug('Writing epoch')
    var = nc.createVariable('epoch', 'i8', dimensions=('time',))
    var.setncattr('long_name', 'Epoch Time')
    var.setncattr('units', 'seconds since 1970-01-01 00:00:00 UTC')
    var[:] = epoch

    logging.debug('Writing hour')
    var = nc.createVariable('hour', 'f8', dimensions=('time',))
    var.setncattr('long_name', 'Hour of Day')
    var.setncattr('units', 'UTC')
    var[:] = hour

    logging.debug('Writing range')
    var = nc.createVariable('range', 'f8', dimensions=('range',))
    var.setncattr('long_name', 'height')
    var.setncattr('units', 'km AGL')
    var[:] = rng

    logging.debug('Writing azimuth')
    var = nc.createVariable('azimuth', 'f8', dimensions=('time', 'range'))
    var.setncattr('long_name', 'Azimuth Angle')
    var.setncattr('units', 'degrees')
    var[:] = np.tile(az, (len(rng), 1)).transpose()

    logging.debug('Writing elevation')
    var = nc.createVariable('elevation', 'f8', dimensions=('time', 'range'))
    var.setncattr('long_name', 'Elevation angle')
    var.setncattr('units', 'degrees above the horizon')
    var[:] = np.tile(elev, (len(rng), 1)).transpose()

    logging.debug('Writing pitch')
    var = nc.createVariable('pitch', 'f8', dimensions=('time', 'range'))
    var.setncattr('long_name', 'Instrument Pitch')
    var.setncattr('units', 'degrees')
    var[:] = np.tile(pitch, (len(rng), 1)).transpose()

    logging.debug('Writing roll')
    var = nc.createVariable('roll', 'f8', dimensions=('time', 'range'))
    var.setncattr('long_name', 'Instrument Roll')
    var.setncattr('units', 'degrees')
    var[:] = np.tile(roll, (len(rng), 1)).transpose()

    logging.debug('Writing velocity')
    var = nc.createVariable('velocity', 'f8', dimensions=('time', 'range'))
    var.setncattr('long_name', 'Doppler velocity')
    var.setncattr('units', 'm/s')
    var.setncattr('comment', 'Positive values are toward the radar')
    var[:] = vel.transpose()

    logging.debug('Writing intensity')
    var = nc.createVariable('intensity', 'f8', dimensions=('time', 'range'))
    var.setncattr('long_name', 'Intensity')
    var.setncattr('units', 'Unitless')
    var.setncattr('comment', 'This is computed as (SNR+1)')
    var[:] = intensity.transpose()

    logging.debug('Writing backscatter')
    var = nc.createVariable('backscatter', 'f8', dimensions=('time', 'range'))
    var.setncattr('long_name', 'Attenuated backscatter')
    var.setncattr('units', 'km^(-1) sr^(-1)')
    var[:] = (beta*1e3).transpose()

    logging.debug('Closing file')
    nc.close()

    logging.info("Netcdf successfully created!")

    return filename


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', dest='in_files', nargs='*')
    parser.add_argument('-o', dest='out_dir')
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true')
    parser.add_argument('-p', '--prefix', dest='prefix', default=None)
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(format='%(asctime)s:%(levelname)s:%(message)s',
                            level=logging.DEBUG)
    else:
        logging.basicConfig(format='%(asctime)s:%(levelname)s:%(message)s',
                            level=logging.INFO)

    for f in args.in_files:
        logging.info('Processing file %s' % f)
        filename = process_file(f, args.out_dir, args.prefix)



import datetime as dt
import pytz
import numpy as np
import numpy.ma as ma
from pathlib import Path
from iss_lidar.consensus_set import ConsensusSet
from iss_lidar.vad import VADSet

datadir = Path(__file__).parent.parent.joinpath("testdata")


def compare_consensus(a: ConsensusSet, b: ConsensusSet):
    """ Compare data from two consensus objects, allowing for small differences
    from saving/reading files, etc"""
    np.testing.assert_allclose(a.alt, b.alt)
    assert isinstance(b.alt, ma.MaskedArray)
    # differences of like 6e-7 in these vals
    np.testing.assert_allclose(a.lat, b.lat)
    assert isinstance(b.lat, ma.MaskedArray)
    np.testing.assert_allclose(a.lon, b.lon)
    assert isinstance(b.lon, ma.MaskedArray)
    np.testing.assert_allclose(a.height, b.height)
    assert isinstance(b.height, ma.MaskedArray)
    np.testing.assert_allclose(a.mean_cnr, b.mean_cnr)
    assert isinstance(b.mean_cnr, ma.MaskedArray)
    assert a.stime == b.stime
    assert isinstance(b.stime, list)
    assert a.etime == b.etime
    assert isinstance(b.etime, list)
    np.testing.assert_allclose(a.u, b.u, equal_nan=True)
    assert isinstance(b.u, ma.MaskedArray)
    np.testing.assert_allclose(a.n_u, b.n_u, equal_nan=True)
    assert isinstance(b.n_u, ma.MaskedArray)
    np.testing.assert_allclose(a.w, b.w, equal_nan=True)
    assert isinstance(b.w, ma.MaskedArray)
    np.testing.assert_allclose(a.n_w, b.n_w, equal_nan=True)
    assert isinstance(b.n_w, ma.MaskedArray)
    np.testing.assert_allclose(a.v, b.v, equal_nan=True)
    assert isinstance(b.v, ma.MaskedArray)
    np.testing.assert_allclose(a.n_v, b.n_v, equal_nan=True)
    assert isinstance(b.n_v, ma.MaskedArray)
    np.testing.assert_allclose(a.speed, b.speed, equal_nan=True)
    assert isinstance(b.speed, ma.MaskedArray)
    np.testing.assert_allclose(a.wdir, b.wdir, equal_nan=True)
    assert isinstance(b.wdir, ma.MaskedArray)
    np.testing.assert_allclose(a.residual, b.residual, equal_nan=True)
    assert isinstance(b.residual, ma.MaskedArray)
    np.testing.assert_allclose(a.correlation, b.correlation, equal_nan=True)
    assert isinstance(b.correlation, ma.MaskedArray)
    assert a.span == b.span
    assert a.window == b.window


def test_create_time_ranges():
    day = dt.date(2021, 6, 30)
    res = ConsensusSet.create_time_ranges(day, dt.timedelta(minutes=30))
    assert len(res) == 48
    assert res[0] == dt.datetime(2021, 6, 30, 00, 00, 00, 00, pytz.UTC)
    assert res[-1] == dt.datetime(2021, 6, 30, 23, 30, 00, 00, pytz.UTC)
    # try hour bins
    res = ConsensusSet.create_time_ranges(day, dt.timedelta(minutes=60))
    assert len(res) == 24
    assert res[0] == dt.datetime(2021, 6, 30, 00, 00, 00, 00, pytz.UTC)
    assert res[1] == dt.datetime(2021, 6, 30, 1, 00, 00, 00, pytz.UTC)
    assert res[-1] == dt.datetime(2021, 6, 30, 23, 00, 00, 00, pytz.UTC)


def test_median_from_consensus_idxs():
    vals = ma.array([1, 2, 3, 4, 5, 4, 3, 2, 1])
    idxs = [0, 1, 2, 7, 8]
    median = ConsensusSet.median_from_consensus_idxs(vals, idxs)
    assert median == ma.median([1, 2, 3, 2, 1])
    # if idxs is empty, return nan without producing numpy warning
    idxs = []
    median = ConsensusSet.median_from_consensus_idxs(vals, idxs)
    assert np.isnan(median)


def test_write_read_netcdf():
    vs = VADSet.from_file(f"{datadir}/VAD_20220501.nc")
    cs = ConsensusSet.from_VADSet(vs, 5, dt.timedelta(minutes=30))
    cs.to_ARM_netcdf(f"{datadir}/test_consensus.nc")
    cs_file = ConsensusSet.from_file(f"{datadir}/test_consensus.nc")
    compare_consensus(cs, cs_file)



import pytest
import json
import netCDF4
import numpy as np
import numpy.ma as ma
import numpy.testing
from pathlib import Path

from iss_lidar.ppi import PPI
from iss_lidar.vad import VAD, VADSet

datadir = Path(__file__).parent.parent.joinpath("testdata")


def assert_allclose(actual, desired, rtol=1e-06, atol=1e-05,
                    equal_nan=True, err_msg='', verbose=True) -> None:
    "Wrap numpy.testing.assert_allclose() with different default tolerances."
    numpy.testing.assert_allclose(actual, desired, rtol, atol,
                                  equal_nan, err_msg, verbose)


def compare_vadsets(a: VADSet, b: VADSet, compare_errors=True):
    """ Compare data from two VADSet objects, allowing for small differences
    from saving/reading files, etc"""
    assert a.min_cnr == b.min_cnr
    assert isinstance(b.min_cnr, int)
    assert_allclose(a.alt, b.alt)
    assert isinstance(b.alt, numpy.ma.MaskedArray)
    # differences of like 6e-7 in these vals
    assert_allclose(a.lat, b.lat)
    assert isinstance(b.lat, numpy.ma.MaskedArray)
    assert_allclose(a.lon, b.lon)
    assert isinstance(b.lon, numpy.ma.MaskedArray)
    assert_allclose(a.height, b.height)
    assert isinstance(b.height, numpy.ma.MaskedArray)
    assert_allclose(a.mean_cnr, b.mean_cnr)
    assert isinstance(b.mean_cnr, numpy.ndarray)
    assert a.stime == b.stime
    assert isinstance(b.stime, list)
    assert a.etime == b.etime
    assert isinstance(b.etime, list)
    assert_allclose(a.el, b.el)
    assert isinstance(b.el, numpy.ma.MaskedArray)
    assert_allclose(a.nbeams, b.nbeams)
    assert isinstance(b.nbeams, numpy.ma.MaskedArray)
    assert_allclose(a.u, b.u, equal_nan=True)
    assert isinstance(b.u, numpy.ndarray)
    if compare_errors:
        assert_allclose(a.du, b.du, equal_nan=True)
        assert isinstance(b.du, numpy.ndarray)
    assert_allclose(a.w, b.w, equal_nan=True)
    assert isinstance(b.w, numpy.ndarray)
    if compare_errors:
        assert_allclose(a.dw, b.dw, equal_nan=True)
        assert isinstance(b.dw, numpy.ndarray)
    assert_allclose(a.v, b.v, equal_nan=True)
    assert isinstance(b.v, numpy.ndarray)
    if compare_errors:
        assert_allclose(a.dv, b.dv, equal_nan=True)
        assert isinstance(b.dv, numpy.ndarray)
    assert_allclose(a.speed, b.speed, equal_nan=True)
    assert isinstance(b.speed, numpy.ndarray)
    assert_allclose(a.wdir, b.wdir, equal_nan=True)
    assert isinstance(b.wdir, numpy.ndarray)
    assert_allclose(a.residual, b.residual, equal_nan=True)
    assert isinstance(b.residual, numpy.ndarray)
    assert_allclose(a.correlation, b.correlation, equal_nan=True)
    assert isinstance(b.correlation, numpy.ndarray)


@pytest.fixture
def ppi():
    """ Read in PPI file """
    ppi = PPI.from_file(
        f"{datadir}/cfrad.20210630_152022_WLS200s-181_133_PPI_50m.nc")
    # use default threshold val in ppi_scans_to_vad of -22
    ppi.threshold_cnr(-22)
    return ppi


@pytest.fixture
def ppis():
    """ Read in 3 PPI scans """
    a = PPI.from_file(
        f"{datadir}/cfrad.20210630_152022_WLS200s-181_133_PPI_50m.nc")
    a.threshold_cnr(-22)
    b = PPI.from_file(
        f"{datadir}/cfrad.20210630_171644_WLS200s-181_133_PPI_50m.nc")
    b.threshold_cnr(-22)
    c = PPI.from_file(
        f"{datadir}/cfrad.20210630_174238_WLS200s-181_133_PPI_50m.nc")
    c.threshold_cnr(-22)
    return(a, b, c)


@pytest.fixture
def locations():
    locs = json.load(open(f"{datadir}/vad/locations.json", "r"))
    x = locs["x"]
    y = locs["y"]
    z = locs["z"]
    return(x, y, z)


@pytest.fixture
def final_vad_winds():
    winds = json.load(open(f"{datadir}/vad/winds.json", "r"))
    u = winds["u"]
    v = winds["v"]
    w = winds["w"]
    return (u, v, w)


@pytest.fixture
def final_vad_errs():
    errs = json.load(open(f"{datadir}/vad/wind_errors.json", "r"))
    du = errs["du"]
    dv = errs["dv"]
    dw = errs["dw"]
    return (du, dv, dw)


@pytest.fixture
def derived_products():
    prods = json.load(open(f"{datadir}/vad/derived_products.json", "r"))
    speed = prods["spd"]
    wdir = prods["dir"]
    res = prods["res"]
    cor = prods["cor"]
    return (speed, wdir, res, cor)


def test_xyz(ppi, locations):
    saved_x, saved_y, saved_z = locations
    x, y, z = VAD.xyz(ppi.ranges, ppi.elevation, ppi.azimuth)
    assert_allclose(x, saved_x)
    assert_allclose(y, saved_y)
    assert_allclose(z, saved_z)


def test_non_nan_idxs(ppi):
    saved_idxs = json.load(open(f"{datadir}/vad/non_nan_idxs.json", "r"))
    idxs = VAD.non_nan_idxs(ppi.vr, 0)
    assert_allclose(idxs, saved_idxs["0"])
    idxs = VAD.non_nan_idxs(ppi.vr, 1)
    assert_allclose(idxs, saved_idxs["1"])
    idxs = VAD.non_nan_idxs(ppi.vr, 2)
    assert_allclose(idxs, saved_idxs["2"])
    idxs = VAD.non_nan_idxs(ppi.vr, 3)
    assert_allclose(idxs, saved_idxs["3"])


def test_calc_A(ppi):
    saved_A = json.load(open(f"{datadir}/vad/a.json", "r"))
    A = VAD.calc_A(ppi.elevation, ppi.azimuth, VAD.non_nan_idxs(ppi.vr, 0))
    assert_allclose(A, saved_A["a0"])
    A = VAD.calc_A(ppi.elevation, ppi.azimuth, VAD.non_nan_idxs(ppi.vr, 1))
    assert_allclose(A, saved_A["a1"])
    A = VAD.calc_A(ppi.elevation, ppi.azimuth, VAD.non_nan_idxs(ppi.vr, 2))
    assert_allclose(A, saved_A["a2"])
    A = VAD.calc_A(ppi.elevation, ppi.azimuth, VAD.non_nan_idxs(ppi.vr, 3))
    assert_allclose(A, saved_A["a3"])


def test_calc_b(ppi):
    saved_b = json.load(open(f"{datadir}/vad/b.json", "r"))
    b = VAD.calc_b(ppi.elevation, ppi.azimuth, ppi.vr,
                   VAD.non_nan_idxs(ppi.vr, 0), 0)
    assert_allclose(b, saved_b["b0"])
    b = VAD.calc_b(ppi.elevation, ppi.azimuth, ppi.vr,
                   VAD.non_nan_idxs(ppi.vr, 1), 1)
    assert_allclose(b, saved_b["b1"])
    b = VAD.calc_b(ppi.elevation, ppi.azimuth, ppi.vr,
                   VAD.non_nan_idxs(ppi.vr, 2), 2)
    assert_allclose(b, saved_b["b2"])
    b = VAD.calc_b(ppi.elevation, ppi.azimuth, ppi.vr,
                   VAD.non_nan_idxs(ppi.vr, 3), 3)
    assert_allclose(b, saved_b["b3"])


def test_arm_vad(ppi, final_vad_winds, final_vad_errs, derived_products):
    vad = VAD.calculate_ARM_VAD(ppi)
    saved_u, saved_v, saved_w = final_vad_winds
    assert_allclose(vad.u, saved_u, equal_nan=True)
    assert_allclose(vad.v, saved_v, equal_nan=True)
    assert_allclose(vad.w, saved_w, equal_nan=True)
    saved_du, saved_dv, saved_dw = final_vad_errs
    assert_allclose(vad.du, saved_du, equal_nan=True)
    assert_allclose(vad.dv, saved_dv, equal_nan=True)
    assert_allclose(vad.dw, saved_dw, equal_nan=True)
    saved_wspd, saved_wdir, saved_res, saved_cor = derived_products
    assert_allclose(vad.speed, saved_wspd, equal_nan=True)
    assert_allclose(vad.wdir, saved_wdir, equal_nan=True)
    assert_allclose(vad.residual, saved_res, equal_nan=True)
    assert_allclose(vad.correlation, saved_cor, equal_nan=True)


def test_vadset_netcdf(ppis):
    """ Compare data and data types between VADSet from VADs and VADSet from
    netcdf """
    vads = []
    for p in ppis:
        vads.append(VAD.calculate_ARM_VAD(p))
    vs = VADSet.from_VADs(vads, -22)
    vs.to_ARM_netcdf(f"{datadir}/test_vadset.nc")
    f = VADSet.from_file(f"{datadir}/test_vadset.nc")
    # vadset from file should match original vadset
    # skip comparing du,dv,dw which are not saved to netcdf currently
    compare_vadsets(vs, f, compare_errors=False)


def test_vadset_from_PPIs(ppis):
    """ Test that vadset from_PPIs function produces the same result as
    calculating VADs individually and then creating VADSet from them """
    vads = []
    for p in ppis:
        vads.append(VAD.calculate_ARM_VAD(p))
    fromVADs = VADSet.from_VADs(vads, -22)
    files = [f"{datadir}/cfrad.20210630_152022_WLS200s-181_133_PPI_50m.nc",
             f"{datadir}/cfrad.20210630_171644_WLS200s-181_133_PPI_50m.nc",
             f"{datadir}/cfrad.20210630_174238_WLS200s-181_133_PPI_50m.nc"]
    fromPPIs = VADSet.from_PPIs(files, -22)
    assert fromVADs == fromPPIs


def test_vad_vs_vadset(ppi):
    # try to see if i get the same netcdf from a single VAD as I get from a
    # VADSet with one VAD in it
    single = VAD.calculate_ARM_VAD(ppi)
    single.to_ARM_nc(f"{datadir}/test_single_vad.nc")
    set = VADSet.from_PPIs(
        [f"{datadir}/cfrad.20210630_152022_WLS200s-181_133_PPI_50m.nc"], -22)
    set.to_ARM_netcdf(f"{datadir}/test_single_vadset.nc")
    a = VADSet.from_file(f"{datadir}/test_single_vad.nc")
    b = VADSet.from_file(f"{datadir}/test_single_vadset.nc")
    compare_vadsets(a, b)


def test_missing_val_if_nan():
    assert VADSet.missing_val_if_nan(25.3) == 25.3
    assert VADSet.missing_val_if_nan(None) == -9999.0
    assert VADSet.missing_val_if_nan(np.nan) == -9999.0
    assert VADSet.missing_val_if_nan(np.nan, -999.0) == -999.0


def test_get_mask():
    # masks made up from thresholds {"correlation_min": 0.9, "residual_max":
    # 0.7, "mean_snr_max": -29.0} cor/res are much stricter than they will be
    # in real life so i can guarantee they will remove values
    corr_mask = np.array([[False, False, False, False, False, False, False,
                           False, False, False, False, False, False, False,
                           False, False, False, False, False, False, False,
                           False, False, False, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True],
                          [False, False, True, True, False, False, False, True,
                           False, False, False, True, True, True, False, False,
                           False, True, True, False, False, False, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True],
                          [False, False, False, False, False, False, False,
                           True, True, True, True, True, True, True, True,
                           True, False, False, False, False, False, False,
                           False, True, False, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True, True, True, True, True, True, True,
                           True, True]])
    res_mask = np.array([[False, False, False, False, False, False, False,
                          False, False, False, False, False, False, False,
                          False, False, False, False, False, False, False,
                          False, False, False, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True],
                         [False, False, False, False, False, False, False,
                          False, False, False, False, True, True, True, True,
                          False, False, True, False, False, False, False,
                          False, False, False, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True],
                         [False, False, False, False, False, False, False,
                          False, True, False, True, True, True, True, True,
                          False, False, False, False, False, False, False,
                          False, False, False, False, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True]])
    snr_mask = np.array([[False, False, False, False, False, False, False,
                          False, False, False, False, False, False, False,
                          False, False, False, False, False, False, False,
                          False, False, False, False, False, False, False,
                          False, False, False, False, False, False, False,
                          False, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True],
                         [False, False, False, False, False, False, False,
                          False, False, False, False, False, False, False,
                          False, False, False, False, False, False, False,
                          False, False, False, False, False, False, False,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True],
                         [False, False, False, False, False, False, False,
                          False, False, False, False, False, False, False,
                          False, False, False, False, False, False, False,
                          False, False, False, False, False, False, False,
                          False, False, False, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True, True,
                          True, True, True, True, True, True, True, True]])
    vs = VADSet.from_file(f"{datadir}/test_vadset.nc")
    # correlation mask
    vs.thresholds = {"correlation_min": 0.9}
    assert ma.allequal(corr_mask, vs.get_mask())
    # residual mask
    vs.thresholds = {"residual_max": 0.7}
    assert ma.allequal(res_mask, vs.get_mask())
    # snr mask
    vs.thresholds = {"mean_snr_min": -29.0}
    assert ma.allequal(snr_mask, vs.get_mask())
    # all three combined
    vs.thresholds = {"correlation_min": 0.9, "residual_max": 0.7,
                     "mean_snr_min": -29.0}
    combined = ma.mask_or(ma.mask_or(corr_mask, res_mask), snr_mask)
    assert ma.allequal(combined, vs.get_mask())
    # make sure other vals in dict don't break processing
    vs.thresholds = {"correlation_min": 0.9, "residual_max": 0.7,
                     "mean_snr_min": -29.0, "test_value": 0.1}
    assert ma.allequal(combined, vs.get_mask())
    # check that blank mask (all false) returned when no thresholds present
    vs.thresholds = {}
    assert np.all(~vs.get_mask())


def test_apply_mask():
    vs_original = VADSet.from_file(f"{datadir}/test_vadset.nc")
    mask = np.full(vs_original.correlation.shape, False)
    mask[0, 0] = mask[0, 1] = mask[0, 2] = mask[0, 3] = mask[0, 4] = True
    mask[1, 5] = mask[1, 6] = mask[1, 7] = mask[1, 8] = mask[1, 9] = True
    mask[2, 0] = mask[2, 1] = mask[2, 2] = mask[2, 3] = mask[2, 4] = True
    vs_masked = VADSet.from_file(f"{datadir}/test_vadset.nc")
    vs_masked.thresholds = {"correlation_min": 0.9, "residual_max": 0.7,
                            "mean_snr_min": -29.0}
    vs_masked.apply_thresholds(mask=mask)
    arr = np.array([True, True, True, True, True])
    assert ma.allequal((ma.getmaskarray(vs_masked.u))[0, 0:5], arr)
    assert ma.allequal((ma.getmaskarray(vs_masked.u))[1, 5:10], arr)
    assert ma.allequal((ma.getmaskarray(vs_masked.u))[2, 0:5], arr)
    assert ma.allequal((ma.getmaskarray(vs_masked.v))[0, 0:5], arr)
    assert ma.allequal((ma.getmaskarray(vs_masked.w))[0, 0:5], arr)
    assert ma.allequal((ma.getmaskarray(vs_masked.du))[0, 0:5], arr)
    assert ma.allequal((ma.getmaskarray(vs_masked.residual))[0, 0:5], arr)
    assert ma.allequal((ma.getmaskarray(vs_masked.speed))[0, 0:5], arr)
    # check that previous mask still present
    # i know that none of the values i masked out were previously masked
    assert ma.allequal(ma.getmaskarray(vs_masked.u),
                       ma.mask_or(ma.getmaskarray(vs_original.u), mask))


def test_nbeams_fraction_threshold():
    nbeams = [100, 50, 200]
    nbeams_used = np.array([[100, 90, 80, 60, 40],[50, 40,30, 20, 10],
                            [200, 150, 125, 100, 75]])
    expected_mask = np.array([[False, False, False, False, True],
                              [False, False, False, True, True],
                              [False, False, False, False, True]])
    mask = VADSet.get_nbeams_used_mask(nbeams, nbeams_used, 0.5)
    assert ma.allequal(mask, expected_mask)
    vs = VADSet.from_file(f"{datadir}/test_vadset.nc")
    mask = VADSet.get_nbeams_used_mask(vs.nbeams, vs.nbeams_used, 0.5)
    # nbeams in this vadset is always 360. check that no nbeams_used are left
    # if nbeams_used is less than 180
    vs.nbeams_used.mask = ma.mask_or(vs.nbeams_used.mask, mask)
    assert ma.min(vs.nbeams_used) >= 180


def test_thresholds_in_netcdf():
    """ Test including threshold vals as attributes in netcdf. """
    vs = VADSet.from_file(f"{datadir}/test_vadset.nc")
    vs.thresholds = {"correlation_min": 0.9, "residual_max": 0.7,
                     "mean_snr_min": -29.0}
    vs.apply_thresholds()
    vs.to_ARM_netcdf(f"{datadir}/test_vadset_thresholds.nc")
    vs_thresholds = netCDF4.Dataset(f"{datadir}/test_vadset_thresholds.nc",
                                    'r')
    attrs = vs_thresholds.__dict__
    assert "threshold_correlation_min" in attrs
    assert "threshold_residual_max" in attrs
    assert "threshold_mean_snr_min" in attrs
    assert attrs["threshold_correlation_min"] == 0.9
    assert attrs["threshold_residual_max"] == 0.7
    assert attrs["threshold_mean_snr_min"] == -29.0
# updating tests example:
# d = {"du": vad.du.tolist(), "dv": vad.dv.tolist(), "dw": vad.dw.tolist()}
# f = open("testdata/vad/wind_errors.json", "w")
# json.dump(d, f)
# f.close()


#!/opt/local/anaconda3/bin/python
# -*- coding: utf-8 -*-
# pylint: disable=C0103,E1101
"""
Created on Thu Feb  4 11:16:18 2021

@author: jgebauer
"""

import numpy as np
import numpy.ma as ma
import matplotlib.pyplot as plt
import scipy.interpolate
from typing import Tuple, List


class gridded_RHI:
    """
    This is a class created for gridded RHI data
    """

    def __init__(self, field, x, z, dx, offset, grid_el, grid_range, time):
        self.field = np.array(field)
        self.x = np.array(x)
        self.z = np.array(z)
        self.dx = dx
        self.x_offset = offset[0]
        self.z_offset = offset[1]
        self.time = np.array(time)
        self.grid_el = grid_el
        self.grid_range = grid_range


def grid_rhi(field, elevation, ranges, dims, dx, offset=None,
             time=None, missing=None):
    """
    This function will take RHI scans and will put them onto a 2-D cartesian
    grid using linear interpolation
    """
    if len(dims) != 2:
        raise IOError('Dims must be a 2 length tuple')

    if offset is not None:
        if len(offset) != 2:
            raise IOError('If offset is specified it must be 2 length tuple')
    else:
        offset = (0, 0)

    if (time is None) and (len(field.shape) == 2):
        times = 1
        time = [0]
        raw = np.array([field])
        el = np.array([elevation])
    elif (time is None) and (len(field.shape) == 3):
        time = np.arange(field.shape[0])
        el = np.copy(elevation)
        raw = np.copy(field)
        times = len(time)
    else:
        times = len(time)
        raw = np.copy(field)
        el = np.copy(elevation)

    if missing is not None:
        raw[raw == missing] = np.nan

    x = ranges[None, :] * np.cos(np.deg2rad(180-el))[:, :, None] + offset[0]
    z = ranges[None, :] * np.sin(np.deg2rad(el))[:, :, None] + offset[1]

    grid_x, grid_z = np.meshgrid(np.arange(dims[0][0], dims[0][1]+1, dx),
                                 np.arange(dims[1][0], dims[1][1]+1, dx))

    grid_range = np.sqrt((grid_x-offset[0])**2 + (grid_z-offset[1])**2)
    grid_el = 180 - np.rad2deg(np.arctan2(grid_z-offset[1], grid_x-offset[0]))

    grid_field = []
    for i in range(times):
        foo = np.where(~np.isnan(x[i]))

        grid_field.append(scipy.interpolate.griddata((x[i, foo[0], foo[1]],
                                                      z[i, foo[0], foo[1]]),
                                                     raw[i, foo[0], foo[1]],
                                                     (grid_x, grid_z)))

    return gridded_RHI(grid_field, grid_x, grid_z, dx, offset, grid_el,
                       grid_range, time)


def coplanar_analysis(vr1, vr2, el1, el2, az):
    """
    This function calculates a coplanar wind field from two gridded RHIs
    """
    u = np.ones(vr1.shape)*np.nan
    w = np.ones(vr1.shape)*np.nan

    for i in range(vr1.shape[0]):
        for j in range(vr1.shape[1]):

            if (~np.isnan(vr1[i, j])) and (~np.isnan(vr2[i, j])):
                M = np.array([[np.sin(np.deg2rad(az)) *
                               np.cos(np.deg2rad(el1[i, j])),
                               np.sin(np.deg2rad(el1[i, j]))],
                              [np.sin(np.deg2rad(az)) *
                               np.cos(np.deg2rad(el2[i, j])),
                               np.sin(np.deg2rad(el2[i, j]))]])

                U = np.linalg.solve(M, np.array([vr1[i, j], vr2[i, j]]))
                u[i, j] = U[0]
                w[i, j] = U[1]

    return u, w


def vr_variance(field, time, t_avg, axis=0):
    """
    The function calculates the vr-variance from a timeseries of scans
    """
    t_avg = t_avg*60
    start = 0
    yo = np.where(time < (time[0]+t_avg))[0]
    end = yo[-1]+1
    var = []
    time_avg = []
    while end < len(time):
        var.append(np.nanvar(field[start:end, :, :], axis=0))
        time_avg.append(np.nanmean(time[start:end]))
        start = end
        yo = np.where(time < (time[start]+t_avg))[0]
        end = yo[-1]+1

    return np.array(var), np.array(time_avg)


def process_LidarSim_scan(scan, scantype, elevation, azimuth, ranges, time):
    """
    This function will work with LidarSim data
    """
    # import this here b/c it's only needed in this function
    from vad import VAD
    if scantype == 'vad':
        el = np.nanmean(elevation)
        vad = VAD.ARM_VAD(scan, ranges, el, azimuth, time)

        return vad

    else:
        print('Not a valid scan type')
        return np.nan


def rhi_vertical_profile(field, elevation, azimuth, ranges, heights, dz, loc,
                         offset=None, time=None, missing=None):
    """
    This function puts vr observations from a RHI onto a vertical grid at one
    location to be used for virtual tower calculations
    """
    if len(loc) != 2:
        raise IOError('Dims must be a 2 length tuple')

    if offset is not None:
        if len(offset) != 3:
            raise IOError('If offset is specified it must be 3 length tuple')
    else:
        offset = (0, 0, 0)

    if (time is None) and (len(field.shape) == 2):
        times = 1
        time = [0]
        raw = np.array([field])
        el = np.array([elevation])
    elif (time is None) and (len(field.shape) == 3):
        time = np.arange(field.shape[0])
        el = np.copy(elevation)
        raw = np.copy(field)
        times = len(time)
    else:
        times = len(time)
        raw = np.copy(field)
        el = np.copy(elevation)

    if missing is not None:
        raw[raw == missing] = np.nan

    x = (ranges[None, :] * np.cos(np.deg2rad(el))[:, :, None] *
         np.sin(np.deg2rad(azimuth)) + offset[0])
    y = (ranges[None, :] * np.cos(np.deg2rad(el))[:, :, None] *
         np.cos(np.deg2rad(azimuth)) + offset[1])
    z = ranges[None, :] * np.sin(np.deg2rad(el))[:, :, None] + offset[2]

    r = np.sqrt(x**2 + y**2)

    z_interp = np.arange(heights[0], heights[1], dz)

    z_ranges = np.sqrt((loc[0]-offset[0])**2 +
                       (loc[1]-offset[1])**2 + (z_interp-offset[2])**2)
    z_el = np.rad2deg(np.arctan2(z_interp-offset[2],
                                 np.sqrt((loc[0]-offset[0])**2 +
                                         (loc[1]-offset[1])**2)))

    grid_field = []
    grid_x, grid_z = np.meshgrid(np.array(np.sqrt(loc[0]**2 + loc[1]**2)),
                                 z_interp)
    for i in range(times):
        grid_field.append(scipy.interpolate.griddata((r[i].ravel(),
                                                      z[i].ravel()),
                                                     raw[i].ravel(),
                                                     (grid_x, grid_z))[:, 0])

    # no vertical_vr object/function included in Josh's code, ignore error
    return vertical_vr(grid_field, loc[0], loc[1], z_interp,  # noqa: F821
                       dz, offset, z_el, z_ranges, azimuth, time)


def virtual_tower(vr, elevation, azimuth, height, uncertainty=0.45):
    """
    This function calculates the wind components for a virtual towers
    """
    if len(vr) == 2:
        u = []
        v = []
        u_uncertainty = []
        v_uncertainty = []
        el1 = np.deg2rad(elevation[0])
        el2 = np.deg2rad(elevation[1])
        az1 = np.deg2rad(azimuth[0])
        az2 = np.deg2rad(azimuth[1])
        for i in range(len(height)):
            if np.isnan(vr[0][i]) or np.isnan(vr[1][i]):
                u.append(np.nan)
                v.append(np.nan)

                M = np.array([[np.sin(az1)*np.cos(el1[i]), np.cos(az1) *
                               np.cos(el1[i])],
                              [np.sin(az2)*np.cos(el2[i]), np.cos(az2) *
                               np.cos(el2[i])]])
                invM = np.linalg.inv(M)
                temp_u = np.sqrt((invM[0, 0]**2)*(uncertainty**2) +
                                 (invM[0, 1]**2)*(uncertainty**2))
                temp_v = np.sqrt((invM[1, 0]**2)*(uncertainty**2) +
                                 (invM[1, 1]**2)*(uncertainty**2))
                u_uncertainty.append(temp_u)
                v_uncertainty.append(temp_v)

            else:
                M = np.array([[np.sin(az1)*np.cos(el1[i]), np.cos(az1) *
                               np.cos(el1[i])],
                              [np.sin(az2)*np.cos(el2[i]), np.cos(az2) *
                               np.cos(el2[i])]])
                temp = np.linalg.solve(M, np.array([vr[0][i], vr[1][i]]))
                u.append(np.copy(temp[0]))
                v.append(np.copy(temp[1]))

                invM = np.linalg.inv(M)
                temp_u = np.sqrt((invM[0, 0]**2)*(uncertainty**2) +
                                 (invM[0, 1]**2)*(uncertainty**2))
                temp_v = np.sqrt((invM[1, 0]**2)*(uncertainty**2) +
                                 (invM[1, 1]**2)*(uncertainty**2))
                u_uncertainty.append(temp_u)
                v_uncertainty.append(temp_v)

        return np.array([u, v]), np.array([u_uncertainty, v_uncertainty])

    elif len(vr) == 3:
        u = []
        v = []
        w = []
        u_uncertainty = []
        v_uncertainty = []
        w_uncertainty = []
        el1 = np.deg2rad(elevation[0])
        el2 = np.deg2rad(elevation[1])
        el3 = np.deg2rad(elevation[2])
        az1 = np.deg2rad(azimuth[0])
        az2 = np.deg2rad(azimuth[1])
        az3 = np.deg2rad(azimuth[2])
        for i in range(len(height)):
            if np.isnan(vr[0][i]) or np.isnan(vr[1][i]):
                u.append(np.nan)
                v.append(np.nan)

                M = np.array([[np.sin(az1)*np.cos(el1[i]), np.cos(az1) *
                               np.cos(el1[i]), np.sin(el1)],
                              [np.sin(az2)*np.cos(el2[i]), np.cos(az2) *
                               np.cos(el2[i]), np.sin(el2)],
                              [np.sin(az3)*np.cos(el3[i]), np.cos(az3) *
                               np.cos(el3[i]),
                               np.sin(el3)]])
                invM = np.linalg.inv(M)
                temp_u = np.sqrt((invM[0, 0]**2)*(uncertainty**2) +
                                 (invM[0, 1]**2)*(uncertainty**2) +
                                 (invM[0, 2]**2)*(uncertainty**2))
                temp_v = np.sqrt((invM[1, 0]**2)*(uncertainty**2) +
                                 (invM[1, 1]**2)*(uncertainty**2) +
                                 (invM[1, 2]**2)*(uncertainty**2))
                temp_w = np.sqrt((invM[2, 0]**2)*(uncertainty**2) +
                                 (invM[2, 1]**2)*(uncertainty**2) +
                                 (invM[2, 2]**2)*(uncertainty**2))

                u_uncertainty.append(temp_u)
                v_uncertainty.append(temp_v)
                w_uncertainty.append(temp_w)
            else:
                M = np.array([[np.sin(az1)*np.cos(el1[i]), np.cos(az1) *
                               np.cos(el1[i]), np.sin(el1)],
                              [np.sin(az2)*np.cos(el2[i]), np.cos(az2) *
                               np.cos(el2[i]), np.sin(el2)],
                              [np.sin(az3)*np.cos(el3[i]), np.cos(az3) *
                               np.cos(el3[i]),
                               np.sin(el3)]])
                temp = np.linalg.solve(M, np.array([vr[0][i], vr[1][i],
                                                    vr[2][i]]))
                u.append(np.copy(temp[0]))
                v.append(np.copy(temp[1]))
                w.append(np.copy(temp[2]))

                invM = np.linalg.inv(M)
                temp_u = np.sqrt((invM[0, 0]**2)*(uncertainty**2) +
                                 (invM[0, 1]**2)*(uncertainty**2) +
                                 (invM[0, 2]**2)*(uncertainty**2))
                temp_v = np.sqrt((invM[1, 0]**2)*(uncertainty**2) +
                                 (invM[1, 1]**2)*(uncertainty**2) +
                                 (invM[1, 2]**2)*(uncertainty**2))
                temp_w = np.sqrt((invM[2, 0]**2)*(uncertainty**2) +
                                 (invM[2, 1]**2)*(uncertainty**2) +
                                 (invM[2, 2]**2)*(uncertainty**2))

                u_uncertainty.append(temp_u)
                v_uncertainty.append(temp_v)
                w_uncertainty.append(temp_w)

        return np.array([u, v, w]), np.array([u_uncertainty, v_uncertainty,
                                              w_uncertainty])
    else:
        print('Input needs to be a length 2 or 3 tuple')
        return np.nan


def lenshow(x, freq=1, tau_min=3, tau_max=12, plot=False):
    """
    This function performs a Lenshow correction to lidar data and calculates
    vertical velocity variance. This function was originally written by Tyler
    Bell at CIMMS in Norman, OK.

    Reads in a timeseries. Freq is in Hz. Default taus are from avg values from
    Bonin Dissertation (2015) Returns avg w'**2 and avg error'**2
    """
    # Find the perturbation of x
    mean = np.mean(x)
    prime = x - mean
    # Get the autocovariance
    acorr, lags = xcorr(prime, prime)
    # var = np.var(prime)
    acov = acorr  # * var
    # Extract lags > 0
    lags = lags[int(len(lags)/2):] * freq
    acov = acov[int(len(acov)/2):]
    # Define the start and end lags
    lag_start = int(tau_min / freq)
    lag_end = int(tau_max / freq)
    # Fit the structure function
    # don't kow which leastsq fn this is referencing
    p1, _ = leastsq(err_funct, [1, .001],  # noqa: F821
                    args=(lags[lag_start:lag_end],
                    acov[lag_start:lag_end]))
    if plot:
        new_lags = np.arange(tau_min, tau_max)
        plt.plot(lags, acov)
        plt.plot(new_lags, fit_funct(p1, new_lags), 'gX')
        plt.plot(0, fit_funct(p1, 0), 'gX')
        plt.xlim(0, tau_max+20)
        plt.xlabel("Lag [s]")
        plt.ylabel("$M_{11} [m^2s^{-2}$]")
    return p1[0], acov[0] - p1[0]


def lenshow_bonin(x, tau_min=1, tint_first_guess=3, freq=1, max_iter=100,
                  plot=False):
    """
    This is a modified version of the Lenshow correction that adaptively
    selects taus based on the data. This function was originally written by
    Tyler Bell at CIMMS in Norman, OK.
    """
    # Find the perturbation of x
    mean = np.mean(x)
    prime = x - mean
    # Get the autocovariance
    acorr, lags = xcorr(prime, prime)
    # var = np.var(prime)
    acov = acorr  # * var
    # Extract lags > 0
    lags = lags[int(len(lags)/2):] * freq
    acov = acov[int(len(acov)/2):]
    # Define the start and end lags
    lag_start = int(tau_min / freq)
    lag_end = int((tau_min+3) / freq)
    # Fit the structure function
    # fit_funct = lambda p, t: p[0] - p[1]*t**(2./3.)
    # err_funct = lambda p, t, y: fit_funct(p, t) - y
    # Iterate to find t_int
    last_tint = (tau_min+3)
    i = 0
    # again, don't know which leastsq this intends to use
    p1, _ = leastsq(err_funct, [.10, .001],  # noqa: F821
                    args=(lags[lag_start:lag_end],
                    acov[lag_start:lag_end]))
    tint = calc_tint(p1[0], freq, acov, lags)
    while np.abs(last_tint - tint) > 1.:
        if i >= max_iter:
            return None
        else:
            i += 1
            last_tint = tint
        p1, _ = leastsq(err_funct, [.10, .001],  # noqa: F821
                        args=(lags[lag_start:lag_end],
                              acov[lag_start:lag_end]))
        tint = calc_tint(p1[0], freq, acov, lags)
    # Find the time where M11(t) = M11(0)/2
    ind = np.min(np.where(acov <= acov[0]/2))
    # Determine what tau to use
    tau_max = np.min([tint/2., lags[ind]])
    # Do the process
    lag_end = int(tau_max / freq)
    if lag_start+1 >= lag_end:
        lag_end = lag_start + 2
    p1, _ = leastsq(err_funct, [.10, .001],  # noqa: F821
                    args=(lags[lag_start:lag_end],
                          acov[lag_start:lag_end]))
    if plot:
        new_lags = np.arange(tau_min, tau_max)
        plt.plot(lags, acov, 'k')
        plt.plot(new_lags, fit_funct(p1, new_lags), 'rX', label='Adaptive')
        plt.plot(0, fit_funct(p1, 0), 'rX')
        plt.xlim(0, tau_max+20)
        plt.xlabel("Lag [s]")
        plt.ylabel("$M_{11} [m^2s^{-2}$]")
    return p1[0], np.abs(acov[0] - p1[0]), tau_max


def fit_funct(p, t):
    """ Used in lenshow, lenshow_bonin calculations """
    p[0] - p[1]*t**(2./3.)


def err_funct(p, t, y):
    """ Used in lenshow, lenshow_bonin calculations """
    fit_funct(p, t) - y


def calc_tint(var, freq, acov, lags):
    """
    This function is used in the lenshow_bonin function
    """
    ind = np.min(np.where(acov < 0))
    return freq**-1. + 1./var * sum(acov[1:ind] / freq)


def xcorr(y1, y2):
    """
    This function calculates the lag correlation of a time series.
    """
    if len(y1) != len(y2):
        raise ValueError('The lenghts of the inputs should be the same')

    corr = np.correlate(y1, y2, mode='full')
    unbiased_size = np.correlate(np.ones(len(y1)), np.ones(len(y1)),
                                 mode='full')
    corr = corr/unbiased_size

    maxlags = len(y1)-1
    lags = np.arange(-maxlags, maxlags + 1)

    return corr, lags


def consensus_avg(vals: ma.array, window: float) -> Tuple[float, List]:
    """
    This function uses consensus averaging to calculate the mean value of all
    data points that lie within a given window size of each other, where it
    finds the window span with the most points in it and with the smallest
    spread in values. It returns the consensus average and a list of which
    indices of the original array were used in the averaging.
    """
    # make sure nans are masked
    vals = ma.masked_invalid(vals)

    # sort but retain original indices
    sorted_idxs = ma.argsort(vals)
    sorted_vals = ma.sort(vals)

    # if all masked, return
    if sorted_vals.mask.all():
        return (np.nan, [])

    max_num_inds = 0
    final_range = np.inf
    for v in sorted_vals[sorted_vals.mask == False]:  # noqa: E712
        booleans = ma.logical_and(sorted_vals >= v, sorted_vals <= v+window)
        inds = ma.where(booleans)[0]
        num_inds = len(inds)
        val_range = sorted_vals[inds[-1]] - sorted_vals[inds[0]]
        if (num_inds > max_num_inds) or (num_inds == max_num_inds
                                         and val_range < final_range):
            max_num_inds = num_inds
            final_range = val_range
            final_inds = inds

    # if there is no window with more than one point in it, return nan
    if max_num_inds < 2:
        return (np.nan, [])

    final_vals = [sorted_vals[i] for i in final_inds]
    orig_idxs = [sorted_idxs[i] for i in final_inds]
    avg = np.mean(final_vals)
    return (avg, orig_idxs)



# utilities for reading/creating lidar files (no actual calculations)

import os
import netCDF4
import pytz
from datetime import datetime
import numpy as np
from typing import Tuple
import matplotlib
import matplotlib.pyplot as plt


def create_filename(date: datetime, destdir: str, filetype: str,
                    prefix: str = None) -> str:
    fname = filetype + "_"
    if prefix:
        fname += prefix + "_"
    fname += date.strftime('%Y%m%d') + '.nc'
    return os.path.join(destdir, fname)


def read_cfradial(file_path):
    """
    This function reads in cfradial files to variables needed for VAD analysis
    """
    lidar_file = netCDF4.Dataset(file_path, 'r')
    cnr = lidar_file.variables['cnr'][:]
    ranges = lidar_file.variables['range'][:]
    vr = lidar_file.variables['radial_wind_speed'][:]
    azimuth = lidar_file.variables['azimuth'][:]
    elevation = lidar_file.variables['elevation'][0]
    latitude = lidar_file.variables['latitude'][:]
    longitude = lidar_file.variables['longitude'][:]
    altitude = lidar_file.variables['altitude'][:]
    # convert start/end to datetimes so they're easier to use.
    # times in cfradial files are in UTC. timezone is not specified in
    # start_time attribute, but is specified as 'Z' instead of 'UTC' in
    # start_datetime attribute, which datetime won't parse.
    start = datetime.strptime(lidar_file.start_time,
                              "%Y-%m-%d %H:%M:%S.%f").replace(tzinfo=pytz.utc)
    end = datetime.strptime(lidar_file.end_time,
                            "%Y-%m-%d %H:%M:%S.%f").replace(tzinfo=pytz.utc)

    return [cnr, ranges, vr, elevation, azimuth, start, end, latitude,
            longitude, altitude]


def wspd_wdir_from_uv(u: np.ndarray,
                      v: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    # calculate derived products
    speed = np.sqrt(u**2 + v**2)
    wdir = 270 - np.rad2deg(np.arctan2(v, u))
    notnan = ~np.isnan(wdir)
    wdir[notnan] %= 360
    return speed, wdir


def time_height_plot(filepath: str, u_mean: np.ndarray, v_mean: np.ndarray,
                     ranges: np.ndarray, heights: np.ndarray):
    ticklabels = matplotlib.dates.DateFormatter("%H:%M")
    fig, ax = plt.subplots(1, 1, figsize=(10, 8))
    fig.suptitle('Winds starting at %s 00:00:00 for 24 hours'
                 % (ranges[0].strftime("%Y%m%d")))
    ax.set_ylabel('Height (m)')
    ax.set_xlabel('HH:MM UTC')
    ax.set_ylim(0, 1500)
    ax.xaxis.set_major_formatter(ticklabels)
    # make times and heights 2d arrays
    times = np.repeat([np.array(ranges)],
                      u_mean.shape[-1], axis=0).swapaxes(1, 0)
    heights = np.repeat([heights], u_mean.shape[0], axis=0)
    ax.barbs(times, heights, u_mean, v_mean,
             barb_increments=dict(half=2.5, full=5, flag=10))
    plt.savefig(filepath)
    plt.close()


from datetime import datetime
from iss_lidar.vad_to_consensus import create_cns_filename


def test_create_cns_filename():
    # with prefix
    destdir = "/scr/rain-isf/apg/projects/swex_2022/iss1/ds/lidar/consensus_30"
    fname = destdir + "/30min_winds_339_20220329.nc"
    dt = datetime(2022, 3, 29)
    assert create_cns_filename(30, dt, destdir, "339") == fname
    # without prefix
    fname = destdir + "/30min_winds_20220329.nc"
    assert create_cns_filename(30, dt, destdir) == fname
    # different timespan
    fname = destdir + "/60min_winds_20220329.nc"
    assert create_cns_filename(60, dt, destdir) == fname




#!/opt/local/anaconda3/bin/python
# pylint: disable=C0103

"""
Created March 2021
Carol Costanza

Output VAD winds into ARM netCDF format from cfradial format
Program works for either 1 cfradial file or multiple within 1 day
"""
import warnings
import glob
import argparse
import numpy as np
from vad import VADSet
from tools import create_filename
from tools import time_height_plot

warnings.simplefilter("ignore")
np.set_printoptions(threshold=np.inf)


def create_parser():
    parser = argparse.ArgumentParser(description="Generate netCDF of VAD winds"
                                     " from PPI scans")
    parser.add_argument("--min_cnr", default=-22, type=float, help="threshold"
                        " cnr below this value")
    parser.add_argument("--threshold_config", default=None, help="path to "
                        "config file of post-VAD thresholding parameters")
    parser.add_argument("--plot",
                        help="create PNG plot w/ same filename as netcdf",
                        dest="plot", default=False, action='store_true')
    parser.add_argument("destdir", help="directory to save VAD files to")
    parser.add_argument("ppifiles", help="ppi file(s) for input", nargs='+')
    return parser.parse_args()


def select_files(path: str):
    # expand glob expression if it didn't get expanded in the shell
    ppi_files = glob.glob(path)
    return sorted(list(ppi_files))


def save(vadset: VADSet, destdir: str, prefix: str = None):
    final_file_path = create_filename(vadset.stime[0], destdir, "VAD", prefix)
    vadset.to_ARM_netcdf(final_file_path)


def main():
    args = create_parser()
    # try to allow either list of files created by shell glob or expression to
    # pass to python glob
    ppi_scans = args.ppifiles
    if len(args.ppifiles) == 1:
        ppi_scans = select_files(args.ppifiles[0])
    vadset = VADSet.from_PPIs(ppi_scans, args.min_cnr)
    if (args.threshold_config):
        # apply thresholds if a config is present
        vadset.load_thresholds(args.threshold_config)
        vadset.apply_thresholds()
    save(vadset, args.destdir)

    if (args.plot):
        plotpath = create_filename(vadset.stime[0], args.destdir,
                                   "VAD").replace(".nc", ".png")
        time_height_plot(plotpath, vadset.u, vadset.v, vadset.stime,
                         vadset.height)


if __name__ == "__main__":
    main()



from .consensus_set import ConsensusSet
from .vad_to_consensus import create_cns_filename
from .vad import VADSet
from .vad import VAD
from .ppi import PPI
from .Lidar_functions import consensus_avg
from .tools import create_filename
from .tools import wspd_wdir_from_uv
from .tools import time_height_plot



#!/opt/local/anaconda3/bin/python
#
# Created June 2021 Carol Costanza
#
# Output consensus averaged VAD winds into ARM netCDF format from
# ARM netCDF format VADs. Optionally also creates a plot.

import argparse
import numpy as np
import datetime as dt
from iss_lidar.vad import VADSet
from iss_lidar.consensus_set import ConsensusSet
from iss_lidar.tools import create_filename
from iss_lidar.tools import time_height_plot

np.set_printoptions(threshold=np.inf)


def parse_args():
    parser = argparse.ArgumentParser(description="Generate consensus averaged"
                                     "netcdfs from VAD files")
    parser.add_argument("vadfile", help="daily VAD file")
    parser.add_argument("destdir", help="directory to save averaged file to")
    parser.add_argument("--plot",
                        help="create PNG plot w/ same filename as netcdf",
                        dest="plot", default=False, action='store_true')
    parser.add_argument("--timespan", default=30, type=int, help="Size of time"
                        " span to average over, in minutes (default 30)")
    return parser.parse_args()


# def plot(final_path: str, u_mean: np.ndarray, v_mean: np.ndarray,
#          ranges: np.ndarray, heights: np.ndarray):
#     ticklabels = matplotlib.dates.DateFormatter("%H:%M")
#     fig, ax = plt.subplots(1, 1, figsize=(10, 8))
#     fig.suptitle('Winds starting at %s 00:00:00 for 24 hours'
#                  % (ranges[0].strftime("%Y%m%d")))
#     ax.set_ylabel('Height (m)')
#     ax.set_xlabel('HH:MM UTC')
#     ax.set_ylim(0, 1500)
#     ax.xaxis.set_major_formatter(ticklabels)
#     # make times and heights 2d arrays
#     times = np.repeat([np.array(ranges)],
#                       u_mean.shape[-1], axis=0).swapaxes(1, 0)
#     heights = np.repeat([heights], u_mean.shape[0], axis=0)
#     ax.barbs(times, heights, u_mean, v_mean,
#              barb_increments=dict(half=2.5, full=5, flag=10))
#     plt.savefig(final_path.replace(".nc", ".png"))
#     plt.close()


def create_cns_filename(timespan: int, stime: dt.datetime, destdir: str,
                        prefix: str = None) -> str:
    """ Create a method for this so I can use it from iss_catalog. Create a
    string descriptor of the timespan used, then supply it to the general
    create_filename function. """
    span_descriptor = str(timespan) + "min_winds"
    fpath = create_filename(stime, destdir, span_descriptor, prefix)
    return fpath


def main():
    args = parse_args()
    vs = VADSet.from_file(args.vadfile)
    cs = ConsensusSet.from_VADSet(vs, 5, dt.timedelta(minutes=args.timespan))

    fpath = create_cns_filename(args.timespan, cs.stime[0], args.destdir)

    if (args.plot):
        time_height_plot(fpath.replace(".nc", ".png"), cs.u, cs.v, cs.stime, cs.height)

    cs.to_ARM_netcdf(fpath)


if __name__ == "__main__":
    main()



import numpy as np
import numpy.ma as ma
from iss_lidar.Lidar_functions import consensus_avg


def test_consensus_avg():
    # examples taken from one day's worth of SWEX VADs. In this file there are
    # never more than three VADs in any half hour bin, have taken some
    # selections at random
    # should return nan for all masked data
    window = 5
    vals = ma.array([np.nan, np.nan, np.nan])
    avg, idxs = consensus_avg(vals, window)
    assert np.isnan(avg)
    assert not idxs
    vals = ma.masked_all((3,))
    avg, idxs = consensus_avg(vals, window)
    assert np.isnan(avg)
    assert not idxs
    # only two within window
    vals = ma.array([-6.539889, -2.504943, -0.9367407])
    avg, idxs = consensus_avg(vals, window)
    assert avg == -1.72084185
    assert idxs == [1, 2]
    # change the order
    vals = ma.array([-2.504943, -6.539889, -0.9367407])
    avg, idxs = consensus_avg(vals, window)
    assert avg == -1.72084185
    assert idxs == [0, 2]
    # all within same window
    vals = ma.array([0.9532883, 0.78370535, 0.12152061])
    avg, idxs = consensus_avg(vals, window)
    assert avg == 0.6195047533333333
    assert sorted(idxs) == [0, 1, 2]
    # make sure that masked vals are getting correctly ignored
    vals = ma.array([0.9532883, np.nan, 0.78370535, np.nan, 0.12152061])
    avg, idxs = consensus_avg(vals, window)
    assert avg == 0.6195047533333333
    assert sorted(idxs) == [0, 2, 4]
    # make sure that masked vals aren't being used in calculations
    vals = ma.array([0.9532883, 2.0, 0.78370535, 2.0, 0.12152061])
    vals = ma.masked_greater(vals, 1.5)
    avg, idxs = consensus_avg(vals, window)
    assert avg == 0.6195047533333333
    assert sorted(idxs) == [0, 2, 4]
    # should return nan if there is no window with more than one point in it
    vals = ma.array([0.0, 10.0, 20.0])
    avg, idxs = consensus_avg(vals, window)
    assert np.isnan(avg)
    assert not idxs



# pylint: disable=C0103,E1101

import os
import sys
import datetime as dt
import pytz
import numpy as np
import numpy.ma as ma
import netCDF4
import matplotlib.pyplot as plt
from typing import Tuple, List
import json

import iss_lidar.Lidar_functions as Lidar_functions
import iss_lidar.tools as tools
from iss_lidar.ppi import PPI


class VAD:
    """
    This is a class created for VAD data
    """

    @staticmethod
    def xyz(ranges: np.ma.MaskedArray, el: np.ma.MaskedArray,
            az: np.ma.MaskedArray) -> Tuple[np.ma.MaskedArray,
                                            np.ma.MaskedArray,
                                            np.ma.MaskedArray]:
        """ Calculate x, y, and z coordinates from range, elevation, and azimuth
        """
        # [None,:] / [:, None] syntax creates 2d array from 1d range and
        # azimuth
        x = (ranges[None, :] * np.cos(np.radians(el))
             * np.sin(np.radians(az[:, None])))
        y = (ranges[None, :] * np.cos(np.radians(el))
             * np.cos(np.radians(az[:, None])))
        z = ranges * np.sin(np.radians(el))
        return (x, y, z)

    @staticmethod
    def non_nan_idxs(vr: np.ma.MaskedArray, i: int) -> np.ma.MaskedArray:
        """ This variable is used to index azimuths, but I'm not really sure why
        """
        return np.argwhere(~np.isnan(vr[:, i])).flatten()

    @staticmethod
    def calc_A(el: np.ma.MaskedArray, az: np.ma.MaskedArray,
               idxs: np.ndarray) -> np.ndarray:
        """ Calculate contents of A matrix """
        A11 = ((np.cos(np.deg2rad(el))**2)
               * np.sum(np.sin(np.deg2rad(az[idxs]))**2))
        A12 = ((np.cos(np.deg2rad(el))**2)
               * np.sum(np.sin(np.deg2rad(az[idxs]))
                        * np.cos(np.deg2rad(az[idxs]))))
        A13 = ((np.cos(np.deg2rad(el))*np.sin(np.deg2rad(el)))
               * np.sum(np.sin(np.deg2rad(az[idxs]))))
        A22 = ((np.cos(np.deg2rad(el))**2)
               * np.sum(np.cos(np.deg2rad(az[idxs]))**2))
        A23 = ((np.cos(np.deg2rad(el))*np.sin(np.deg2rad(el)))
               * np.sum(np.cos(np.deg2rad(az[idxs]))))
        A33 = len(az[idxs]) * (np.sin(np.deg2rad(el))**2)

        A = ma.array([[A11, A12, A13], [A12, A22, A23], [A13, A23, A33]])
        return A

    @staticmethod
    def nan_if_masked(barray: list) -> list:
        """ Not currently used, but leaving for reference """
        return [b if not np.ma.is_masked(b) else np.nan for b in barray]

    @staticmethod
    def calc_b(el: np.ma.MaskedArray, az: np.ma.MaskedArray,
               vr: np.ma.MaskedArray, idxs: np.ndarray, i: int) -> np.ndarray:
        """ Calculate contents of b matrix """
        # If all of the vr[idxs, i] array is masked, then b1, b2, and b3 will
        # be masked, and the np.array() creation will report a warning about
        # converting a masked element to nan.  The best way I could figure out
        # to avoid that warning was to explicitly convert a masked result.
        b1 = np.cos(np.deg2rad(el)) * np.sum(vr[idxs, i] *
                                             np.sin(np.deg2rad(az[idxs])))
        b2 = np.cos(np.deg2rad(el)) * np.sum(vr[idxs, i] *
                                             np.cos(np.deg2rad(az[idxs])))
        b3 = np.sin(np.deg2rad(el)) * np.sum(vr[idxs, i])

        # now using ma arrays in VAD, so no need to convert to np arrays w/ nan
        # b = np.array(VAD.nan_if_masked([b1, b2, b3]))
        b = ma.array([b1, b2, b3])
        return b

    def __init__(self, ppi: PPI, u: np.ndarray, v: np.ndarray, w: np.ndarray,
                 speed: np.ndarray, wdir: np.ndarray, du: np.ndarray,
                 dv: np.ndarray, dw: np.ndarray, nbeams_used: np.ndarray,
                 z: np.ndarray, residual: np.ndarray, correlation: np.ndarray):
        self.u = ma.masked_invalid(u)
        self.v = ma.masked_invalid(v)
        self.w = ma.masked_invalid(w)
        self.speed = ma.masked_invalid(speed)
        self.wdir = ma.masked_invalid(wdir)
        self.du = ma.masked_invalid(du)
        self.dv = ma.masked_invalid(dv)
        self.dw = ma.masked_invalid(dw)
        self.nbeams_used = ma.masked_invalid(nbeams_used)
        self.z = ma.masked_invalid(z)
        self.residual = ma.masked_invalid(residual)
        self.correlation = ma.masked_invalid(correlation)
        self.stime = ppi.starttime
        self.etime = ppi.endtime
        self.el = ppi.elevation
        self.nbeams = len(ppi.azimuth)
        self.alt = ppi.alt
        self.lat = ppi.lat
        self.lon = ppi.lon
        self.mean_cnr = ppi.mean_cnr()

    @classmethod
    def calculate_ARM_VAD(cls, ppi: PPI, missing=None):
        """
        This function calculates VAD wind profiles using the technique shown in
        Newsom et al. (2019). This function calculates VAD output for a single
        PPI scan.
        vr: 2d array (azimuth x range)
        ranges: 1d array
        el: scalar
        az: 1d array
        """
        # remove missing radial vel data
        if missing is not None:
            ppi.vr[ppi.vr == missing] = np.nan

        # calculate XYZ coordinates of data
        x, y, z = VAD.xyz(ppi.ranges, ppi.elevation, ppi.azimuth)

        u = np.ones(len(ppi.ranges))*np.nan
        v = np.ones(len(ppi.ranges))*np.nan
        w = np.ones(len(ppi.ranges))*np.nan
        du = np.ones(len(ppi.ranges))*np.nan
        dv = np.ones(len(ppi.ranges))*np.nan
        dw = np.ones(len(ppi.ranges))*np.nan
        nbeams_used = np.ones(len(ppi.ranges))*np.nan

        for i in range(len(ppi.ranges)):
            idxs = VAD.non_nan_idxs(ppi.vr, i)
            nbeams_used[i] = len(idxs)

            # need at least 25% of the azimuth radial velocities available
            if nbeams_used[i] <= len(ppi.azimuth)/4 or np.isnan(nbeams_used[i]):
                continue

            A = VAD.calc_A(ppi.elevation, ppi.azimuth, idxs)
            invA = np.linalg.inv(A)

            du[i] = np.sqrt(invA[0, 0])
            dv[i] = np.sqrt(invA[1, 1])
            dw[i] = np.sqrt(invA[2, 2])

            b = VAD.calc_b(ppi.elevation, ppi.azimuth, ppi.vr, idxs, i)
            temp = invA.dot(b)

            u[i] = temp[0]
            v[i] = temp[1]
            w[i] = temp[2]

        # calculate derived products
        speed, wdir = tools.wspd_wdir_from_uv(u, v)

        residual = np.sqrt(np.nanmean(((((u*x)+(v*y)+((w*z)
                           [None, :]))/np.sqrt(x**2+y**2+z**2))-ppi.vr)**2,
                                      axis=0))
        u_dot_r = ((u*x)+(v*y)+((w*z)[None, :]))/np.sqrt(x**2+y**2+z**2)
        mean_u_dot_r = np.nanmean(((u*x)+(v*y)+((w*z)[None, :])) /
                                  np.sqrt(x**2+y**2+z**2), axis=0)
        mean_vr = np.nanmean(ppi.vr, axis=0)
        correlation = (np.nanmean((u_dot_r-mean_u_dot_r)*(ppi.vr-mean_vr),
                                  axis=0) /
                       (np.sqrt(np.nanmean((u_dot_r-mean_u_dot_r)**2, axis=0))
                        * np.sqrt(np.nanmean((ppi.vr-mean_vr)**2, axis=0))))

        return cls(ppi, u, v, w, speed, wdir, du, dv, dw, nbeams_used,
                   z, residual, correlation)

    def plot_(self, filename: str, plot_time: int = None, title: str = None):
        """
        This function will plot wind profiles from the VAD object
        """

        if plot_time is None:
            plot_time = len(self.time) - 1
            start = 0
        else:
            start = plot_time

        for i in range(start, plot_time+1):
            plt.figure(figsize=(12, 8))
            ax = plt.subplot(141)
            plt.plot(self.speed[i], self.z)
            ax.set_xlim(0, 30)
            ax.set_ylim([0, np.max(self.z)])
            ax.set_xlabel('Wind Speed [m/s]')
            ax.set_ylabel('Height [m]')

            ax = plt.subplot(142)
            plt.plot(self.wdir[i], self.z)
            ax.set_xlim([0, 360])
            ax.set_ylim([0, np.max(self.z)])
            ax.set_xlabel('Wind Direction')

            ax = plt.subplot(143)
            plt.plot(self.w[i], self.z)
            ax.set_xlim(-5, 5)
            ax.set_ylim([0, np.max(self.z)])
            ax.set_xlabel('Vertical Velocity [m/s]')

            ax = plt.subplot(144)
            plt.plot(self.residual[i], self.z)
            ax.set_xlim(0, 10)
            ax.set_ylim([0, np.max(self.z)])
            ax.set_xlabel('Residual')

            if title is not None:
                plt.title(title)

            plt.tight_layout()

            if (start == 0) and (len(self.time) > 1):
                if os.path.isdir(filename):
                    plt.savefig(filename + str(self.time[i]))
                else:
                    os.mkdir(filename)
                    plt.savefig(filename + str(self.time[i]))
            else:
                plt.savefig(filename)

            plt.close()

    def plot_comp(self, altitude: float, ws: np.ndarray, wd: np.ndarray,
                  w: np.ndarray, z: np.ndarray, wprof_alt: float,
                  filename: str, plot_time: int = None, title: str = None):
        """
        This function will plot wind profiles from the VAD object and
        Comparison to other ws,wd,w,z
        """
        if plot_time is None:
            plot_time = len(self.time) - 1
            start = 0
        else:
            start = plot_time

        for i in range(start, plot_time+1):
            plt.figure(figsize=(12, 8))
            ax = plt.subplot(141)
            plt.plot(self.speed[i], self.z+altitude, label='Lidar VAD')
            if len(ws) > 0:
                plt.plot(ws, z+wprof_alt, color='r', label='Wind Profiler')
            plt.legend(loc='upper right')
            ax.set_xlim(0, 30)
            ax.set_ylim([0, 3000])
            ax.set_xlabel('Wind Speed [m/s]')
            ax.set_ylabel('Height [m]')

            ax = plt.subplot(142)
            plt.plot(self.wdir[i], self.z+altitude, label='Lidar VAD')
            if len(wd) > 0:
                plt.plot(wd, z+wprof_alt, color='r', label='Wind Profiler')
            plt.legend(loc='upper right')
            ax.set_xlim([0, 360])
            ax.set_ylim([0, 3000])
            ax.set_xlabel('Wind Direction')

            ax = plt.subplot(143)
            plt.plot(self.w[i], self.z+altitude, label='Lidar VAD')
            if len(w) > 0:
                plt.plot(w, z+wprof_alt, color='r', label='Wind Profiler')
            plt.legend(loc='upper right')
            ax.set_xlim(-5, 5)
            ax.set_ylim([0, 3000])
            ax.set_xlabel('Vertical Velocity [m/s]')

            ax = plt.subplot(144)
            plt.plot(self.residual[i], self.z+altitude)
            ax.set_xlim(0, 10)
            ax.set_ylim([0, 3000])
            ax.set_xlabel('Residual')

            if title is not None:
                plt.title(title)

            plt.tight_layout()

            if (start == 0) and (len(self.time) > 1):
                if os.path.isdir(filename):
                    plt.savefig(filename + str(self.time[i]))
                else:
                    os.mkdir(filename)
                    plt.savefig(filename + str(self.time[i]))
            else:
                plt.savefig(filename)

            plt.close()

    def to_ARM_nc(self, file_path: str, min_cnr: float = -22):
        """
        Create VAD output in ARM netCDF format
        """
        # really a single VAD should have the same netCDF representation as a
        # VADSet with one vad in it, so use that code to generate this netCDF
        vs = VADSet.from_VADs([self], min_cnr)
        vs.to_ARM_netcdf(file_path)


class VADSet:
    """ Class to hold data from a series of VAD calculations """

    @staticmethod
    def missing_val_if_nan(val: float, missing_val: float = -9999.0) -> float:
        """ netCDF4 library automatically sets nans to missing in writing to
        array variables, but not scalars. """
        if val is None or np.isnan(val):
            return missing_val
        return val

    def __init__(self, mean_cnr: np.ndarray, min_cnr: int,
                 alt: np.ma.MaskedArray, lat: np.ma.MaskedArray,
                 lon: np.ma.MaskedArray, height: np.ma.MaskedArray,
                 stime: list, etime: list, el: list, nbeams: list,
                 u: np.ndarray, du: np.ndarray, v: np.ndarray,
                 dv: np.ndarray, w: np.ndarray, dw: np.ndarray,
                 nbeams_used: np.ndarray, speed: np.ndarray,
                 wdir: np.ndarray, residual: np.ndarray,
                 correlation: np.ndarray):
        self.mean_cnr = mean_cnr
        self.min_cnr = min_cnr
        self.alt = alt
        self.lat = lat
        self.lon = lon
        self.height = height
        self.stime = stime
        self.etime = etime
        self.el = el
        self.nbeams = nbeams
        self.u = u
        self.v = v
        self.w = w
        self.du = du
        self.dv = dv
        self.dw = dw
        self.nbeams_used = nbeams_used
        self.speed = speed
        self.wdir = wdir
        self.residual = residual
        self.correlation = correlation
        # use load_thresholds to add thresholds dict
        self.thresholds = None

    @classmethod
    def from_VADs(cls, vads: List[VAD], min_cnr: int):
        # avoid using the heights from a VAD with bad heights
        i = 0
        while (vads[i].z[0] == 0 and i < len(vads)):
            i += 1
        z = vads[i].z
        return cls(ma.array([i.mean_cnr for i in vads]),
                   min_cnr,
                   # use any vad for location, presumably it doesn't change
                   vads[0].alt,
                   vads[0].lat,
                   vads[0].lon,
                   # currently we assume that heights are the same for all VAD
                   # in a set
                   z,
                   # lists of datetime objects, tz-aware
                   [i.stime for i in vads],
                   [i.etime for i in vads],
                   # data from VAD objects
                   [i.el for i in vads],
                   ma.array([i.nbeams for i in vads]),
                   ma.array([i.u for i in vads]),
                   ma.array([i.du for i in vads]),
                   ma.array([i.v for i in vads]),
                   ma.array([i.dv for i in vads]),
                   ma.array([i.w for i in vads]),
                   ma.array([i.dw for i in vads]),
                   ma.array([i.nbeams_used for i in vads]),
                   ma.array([i.speed for i in vads]),
                   ma.array([i.wdir for i in vads]),
                   ma.array([i.residual for i in vads]),
                   ma.array([i.correlation for i in vads]))

    @classmethod
    def from_file(cls, filename: str):
        """ Create a VADSet object from a daily VAD netcdf file """
        f = netCDF4.Dataset(filename, 'r')
        stime = list(netCDF4.num2date(f.variables['time'][:],
                                      f.variables['time'].units,
                                      only_use_python_datetimes=True,
                                      only_use_cftime_datetimes=False))
        # reconstitute endtime from scan duration
        etime = list(netCDF4.num2date(f.variables['time'][:] +
                                      f.variables['scan_duration'],
                                      f.variables['time'].units,
                                      only_use_python_datetimes=True,
                                      only_use_cftime_datetimes=False))
        # make dates tz aware
        stime = [s.replace(tzinfo=pytz.utc) for s in stime]
        etime = [e.replace(tzinfo=pytz.utc) for e in etime]
        # scalar values with a missing_value get read in as np.ma.MaskedArray
        # if the single value is not masked, but as np.ma.core.MaskedConstant
        # if the single value is masked. Instances of MaskedConstant don't
        # compare well to each other or to other data types, so if the scalar
        # lat/lon/alt is masked, replace it with a null-dimensioned masked
        # array with a value of nan.
        alt = (ma.array(np.nan) if f.variables['alt'][:] is np.ma.masked
               else f.variables['alt'][:])
        lat = (ma.array(np.nan) if f.variables['lat'][:] is np.ma.masked
               else f.variables['lat'][:])
        lon = (ma.array(np.nan) if f.variables['lon'][:] is np.ma.masked
               else f.variables['lon'][:])
        min_cnr = None
        try:  # accept min_cnr as either an attribute or a variable
            min_cnr = int(f.input_ppi_min_cnr)
        except AttributeError:
            min_cnr = f.variables['snr_threshold'][:]
        # create blank du, dv, dw if not present in netcdf
        du = np.full(f.variables['u'][:].shape, np.nan)
        dv = np.full(f.variables['v'][:].shape, np.nan)
        dw = np.full(f.variables['w'][:].shape, np.nan)
        try:
            du = f.variables['u_error'][:]
            dv = f.variables['v_error'][:]
            dw = f.variables['w_error'][:]
        except KeyError:
            pass
        return cls(ma.array(f.variables['mean_snr'][:]),
                   min_cnr,
                   alt,
                   lat,
                   lon,
                   f.variables['height'][:],
                   stime,
                   etime,
                   ma.array(f.variables['elevation_angle'][:]),
                   ma.array(f.variables['nbeams'][:]),
                   ma.array(f.variables['u'][:]),
                   ma.array(du),
                   ma.array(f.variables['v'][:]),
                   ma.array(dv),
                   ma.array(f.variables['w'][:]),
                   ma.array(dw),
                   ma.array(f.variables['nbeams_used'][:]),
                   ma.array(f.variables['wind_speed'][:]),
                   ma.array(f.variables['wind_direction'][:]),
                   ma.array(f.variables['residual'][:]),
                   ma.array(f.variables['correlation'][:]))

    @classmethod
    def from_PPIs(cls, ppi_files: List[str], min_cnr: int):
        """ Create a VADSet object by performing VAD calculation on each of a
        list of PPI files """
        vads = []
        for f in ppi_files:
            ppi = PPI.from_file(f)

            # for low elevation angles, VAD output isn't very helpful
            if ppi.elevation < 6:
                continue
            # need at least 3 different azimuths to calculate VAD winds
            if len(ppi.azimuth) < 3:
                continue
            ppi.threshold_cnr(min_cnr)

            # generate VAD for this timestep
            vad = VAD.calculate_ARM_VAD(ppi)
            vads.append(vad)
        if not vads:
            # didn't successfully create any vads. can't continue processing.
            return

        return cls.from_VADs(vads, min_cnr)

    def __eq__(self, other):
        # for testing.
        if (np.array_equal(self.mean_cnr, other.mean_cnr)
            and np.array_equal(self.min_cnr, other.min_cnr)
            and np.array_equal(self.alt, other.alt, equal_nan=True)
            and np.array_equal(self.lat, other.lat, equal_nan=True)
            and np.array_equal(self.lon, other.lon, equal_nan=True)
            and np.array_equal(self.height, other.height, equal_nan=True)
            and np.array_equal(self.stime, other.stime)
            and np.array_equal(self.etime, other.etime)
            and np.array_equal(self.el, other.el, equal_nan=True)
            and np.array_equal(self.nbeams, other.nbeams, equal_nan=True)
            and np.array_equal(self.u, other.u, equal_nan=True)
            and np.array_equal(self.v, other.v, equal_nan=True)
            and np.array_equal(self.w, other.w, equal_nan=True)
            and np.array_equal(self.du, other.du, equal_nan=True)
            and np.array_equal(self.dv, other.dv, equal_nan=True)
            and np.array_equal(self.dw, other.dw, equal_nan=True)
            and np.array_equal(self.nbeams_used, other.nbeams_used,
                               equal_nan=True)
            and np.array_equal(self.speed, other.speed, equal_nan=True)
            and np.array_equal(self.wdir, other.wdir, equal_nan=True)
            and np.array_equal(self.residual, other.residual, equal_nan=True)
            and np.array_equal(self.correlation, other.correlation,
                               equal_nan=True)):
            return True
        return False

    def load_thresholds(self, config: str):
        """ Read dict of thresholding values from the given json file path. """
        # sample threshold config contents:
        # {
        #     "correlation_min": 0.3,
        #     "residual_max": 10.0,
        #     "mean_snr_min": -29.0,
        #     "nbeams_used_min_fraction": 0.5
        # }
        # note: mean_snr name used here instead of min_cnr so it will match
        # what is in netcdf input/output. really the quantity is cnr, and in
        # the VAD objects it's correctly named mean_cnr
        self.thresholds = json.load(open(config, "r"))

    @staticmethod
    def get_nbeams_used_mask(nbeams: List, nbeams_used: np.array,
                             threshold: float) -> np.array:
        """ Since total nbeams can vary between scans, we give the threshold as
        a fraction of possible nbeams that should be present. """
        # separate out this function for testing
        # nbeams_used = time x height array
        # nbeams is just list of length times. broadcast it to an array of
        # time x height
        nbeams = ma.repeat(ma.array([nbeams]), nbeams_used.shape[-1],
                           axis=0).swapaxes(1, 0)
        fraction = nbeams_used / nbeams
        fraction = ma.masked_where(fraction < threshold,
                                   fraction)
        return ma.getmaskarray(fraction)

    def get_mask(self) -> np.array:
        """ Go through all variables to threshold on, combine into a single
        mask that can be applied to all variables. Return the mask, in the form
        of a boolean numpy array where true indicates the value is masked"""
        masks = []
        # create base mask of all false (so, will not mask any values) assuming
        # all time-height arrays will have same dimensions, so just use
        # correlation to get their shape
        final_mask = np.full(self.correlation.shape, False)
        if "correlation_min" in self.thresholds:
            tmp = ma.masked_where(self.correlation
                                  < self.thresholds["correlation_min"],
                                  self.correlation)
            masks.append(ma.getmaskarray(tmp))
        if "residual_max" in self.thresholds:
            tmp = ma.masked_where(self.residual
                                  > self.thresholds["residual_max"],
                                  self.residual)
            masks.append(ma.getmaskarray(tmp))
        if "mean_snr_min" in self.thresholds:
            tmp = ma.masked_where(self.mean_cnr
                                  < self.thresholds["mean_snr_min"],
                                  self.mean_cnr)
            masks.append(ma.getmaskarray(tmp))
        if "nbeams_used_min_fraction" in self.thresholds:
            mask = VADSet.get_nbeams_used_mask(self.nbeams, self.nbeams_used,
                                               self.thresholds["nbeams_used_mi"
                                                               "n_fraction"])
            masks.append(mask)

        # combine all masks
        for m in masks:
            final_mask = ma.mask_or(final_mask, m)
        return final_mask

    def apply_thresholds(self, mask: np.array = None):
        """ Apply thresholding for max/min values of different variables, if
        present. Option to supply a mask is for testing, shoud otherwise use
        get_mask() to mask by thresholds"""
        if self.thresholds is None:
            # can't apply thresholds if they're not present in the object.
            return
        if mask is None:
            mask = self.get_mask()
        # apply mask to values. use mask_or to preserve existing masked vals.
        # wind components
        self.u.mask = ma.mask_or(ma.getmaskarray(self.u), mask)
        self.v.mask = ma.mask_or(ma.getmaskarray(self.v), mask)
        self.w.mask = ma.mask_or(ma.getmaskarray(self.w), mask)
        self.du.mask = ma.mask_or(ma.getmaskarray(self.du), mask)
        self.dv.mask = ma.mask_or(ma.getmaskarray(self.dv), mask)
        self.dw.mask = ma.mask_or(ma.getmaskarray(self.dw), mask)
        self.speed.mask = ma.mask_or(ma.getmaskarray(self.speed), mask)
        self.wdir.mask = ma.mask_or(ma.getmaskarray(self.wdir), mask)
        # ancillary variables
        self.nbeams_used.mask = ma.mask_or(ma.getmaskarray(self.nbeams_used),
                                           mask)
        self.residual.mask = ma.mask_or(ma.getmaskarray(self.residual), mask)
        self.correlation.mask = ma.mask_or(ma.getmaskarray(self.correlation),
                                           mask)
        self.mean_cnr.mask = ma.mask_or(ma.getmaskarray(self.mean_cnr), mask)

    def consensus_average(self, ranges: list) -> Tuple[ma.array, ma.array,
                                                       ma.array]:
        """ Return consensus averaged u,v,w for 30-min increments starting at
        list of time ranges """
        u_mean = np.zeros((len(ranges), len(self.height)))
        v_mean = np.zeros((len(ranges), len(self.height)))
        w_mean = np.zeros((len(ranges), len(self.height)))

        # for ind_start in range(len(secs)-1):
        for idx, r in enumerate(ranges):
            start = r
            end = start + dt.timedelta(minutes=30)
            thirty_min_ind = [i for i in range(len(self.stime))
                              if self.stime[i] >= start and
                              self.stime[i] < end]

            if len(thirty_min_ind) == 0:
                u_mean[idx, :] = np.nan
                v_mean[idx, :] = np.nan
                w_mean[idx, :] = np.nan
                continue

            u_all = np.array([self.u[i] for i in thirty_min_ind])
            v_all = np.array([self.v[i] for i in thirty_min_ind])
            w_all = np.array([self.w[i] for i in thirty_min_ind])

            for hgt in range(len(self.height)):
                # run consensus averaging with a window of 5 m/s
                u_mean[idx, hgt] = Lidar_functions.consensus_avg(u_all[:, hgt],
                                                                 5)
                v_mean[idx, hgt] = Lidar_functions.consensus_avg(v_all[:, hgt],
                                                                 5)
                w_mean[idx, hgt] = Lidar_functions.consensus_avg(w_all[:, hgt],
                                                                 5)
        return (ma.masked_invalid(u_mean), ma.masked_invalid(v_mean),
                ma.masked_invalid(w_mean))

    def to_ARM_netcdf(self, filepath: str):
        str_start_time = self.stime[0].strftime('%Y-%m-%d %H:%M:%S %Z')
        str_day_start_time = self.stime[0].strftime('%Y-%m-%d')
        # create netcdf file
        nc_file = netCDF4.Dataset(filepath, 'w', format='NETCDF4')
        # create dimensions: time, height, bound
        nc_file.createDimension('time', None)
        # still currently assuming that all files in a VADSet have the same
        # heights
        nc_file.createDimension('height', len(self.height))
        nc_file.createDimension('bound', 2)
        # create time variables
        base_time = nc_file.createVariable('base_time', 'i')
        base_time.string = str_start_time
        base_time.long_name = 'Base time in Epoch'
        base_time.units = 'seconds since 1970-01-01 00:00:00 UTC'
        base_time.ancillary_variables = 'time_offset'
        base_time[:] = netCDF4.date2num(self.stime[0], base_time.units)
        time_offset = nc_file.createVariable('time_offset', 'd', 'time')
        time_offset.long_name = 'Time offset from base_time'
        time_offset.units = 'seconds since ' + str_start_time
        time_offset.ancillary_variables = "base_time"
        time_offset[:] = netCDF4.date2num(self.stime, time_offset.units)
        stimes = nc_file.createVariable('time', 'd', 'time')
        stimes.long_name = 'Time offset from midnight'
        stimes.units = 'seconds since ' + str_day_start_time + ' 00:00:00 UTC'
        stimes.bounds = 'time_bounds'
        stimes[:] = netCDF4.date2num(self.stime, stimes.units)
        time_bounds = nc_file.createVariable('time_bounds', 'd',
                                             ('time', 'bound'))
        time_bounds[:, :] = list(zip(netCDF4.date2num(self.stime,
                                                      base_time.units),
                                     netCDF4.date2num(self.etime,
                                                      base_time.units)))
        # create height variable
        height = nc_file.createVariable('height', 'f', 'height')
        height[:] = self.height
        height.long_name = 'Height above ground level'
        height.units = 'm'
        height.standard_name = 'height'

        # add base vars
        self.add_base_variables(nc_file)
        self.add_aux_variables(nc_file)

        # add basic attributes for all file types
        nc_file.Conventions = 'ARM-1.1'
        nc_file.history = 'created on ' + dt.datetime.utcnow().strftime(
            '%Y/%m/%d %H:%M:%S UTC')
        nc_file.command_line = " ".join(sys.argv)
        # add file type specific attributes
        self.add_attributes(nc_file)

        nc_file.close()

    def add_base_variables(self, nc_file: netCDF4.Dataset):
        """
        Add base wind variables to netCDF
        """
        # location
        lat = nc_file.createVariable('lat', 'f')
        lat.missing_value = -9999.0
        lat[:] = VADSet.missing_val_if_nan(self.lat)
        lat.long_name = 'North latitude'
        lat.units = 'degree_N'
        lat.valid_min = -90
        lat.valid_max = 90
        lat.standard_name = 'latitude'
        lon = nc_file.createVariable('lon', 'f')
        lon.missing_value = -9999.0
        lon[:] = VADSet.missing_val_if_nan(self.lon)
        lon.long_name = 'East longitude'
        lon.units = 'degree_E'
        lon.valid_min = -180
        lon.valid_max = 180
        lon.standard_name = 'longitude'
        alt = nc_file.createVariable('alt', 'f')
        alt.missing_value = -9999.0
        alt[:] = VADSet.missing_val_if_nan(self.alt)
        alt.long_name = 'Altitude above mean sea level'
        alt.units = 'm'
        alt.standard_name = 'altitude'
        # uvw
        u = nc_file.createVariable('u', 'f', ('time', 'height'))
        u.missing_value = -9999.0
        u[:, :] = self.u
        u.long_name = 'Eastward component of wind vector'
        u.units = 'm/s'
        v = nc_file.createVariable('v', 'f', ('time', 'height'))
        v.missing_value = -9999.0
        v[:, :] = self.v
        v.long_name = 'Northward component of wind vector'
        v.units = 'm/s'
        w = nc_file.createVariable('w', 'f', ('time', 'height'))
        w.missing_value = -9999.0
        w[:, :] = self.w
        w.long_name = 'Vertical component of wind vector'
        w.units = 'm/s'
        # wspd/wdir
        wind_speed = nc_file.createVariable('wind_speed', 'f',
                                            ('time', 'height'))
        wind_speed.missing_value = -9999.0
        wind_speed[:, :] = self.speed
        wind_speed.long_name = 'Wind speed'
        wind_speed.units = 'm/s'
        wind_direction = nc_file.createVariable('wind_direction', 'f',
                                                ('time', 'height'))
        wind_direction.missing_value = -9999.0
        wind_direction[:, :] = self.wdir
        wind_direction.long_name = 'Wind direction'
        wind_direction.units = 'degree'
        # residual, correlation, mean_snr (present in both vad + consensus)
        residual = nc_file.createVariable('residual', 'f', ('time', 'height'))
        residual.missing_value = -9999.0
        residual[:, :] = self.residual
        residual.long_name = 'Fit residual'
        residual.units = 'm/s'
        correlation = nc_file.createVariable('correlation', 'f',
                                             ('time', 'height'))
        correlation.missing_value = -9999.0
        correlation[:, :] = self.correlation
        correlation.long_name = 'Fit correlation coefficient'
        correlation.units = 'unitless'
        # put mean cnr in file as mean snr to match ARM format (it is still cnr
        # though)
        mean_snr = nc_file.createVariable('mean_snr', 'f',
                                          ('time', 'height'))
        mean_snr.missing_value = -9999.0
        mean_snr[:, :] = self.mean_cnr
        mean_snr.long_name = ('Signal to noise ratio averaged over nbeams '
                              '(derived from CNR)')
        mean_snr.units = 'unitless'

    def add_aux_variables(self, nc_file: netCDF4.Dataset):
        """
        Add auxiliary variables (generally metadata, that are different between
        VADSet and ConsensusSet)
        """
        # Currently not used because values are incorrect
        # Need to relook at the equations in calcA() 
        # uvw errors
        #u_error = nc_file.createVariable('u_error', 'f', ('time', 'height'))
        #u_error.missing_value = -9999.0
        #u_error[:, :] = self.du
        #u_error.long_name = ('Sampling uncertainty in eastward component of'
        #                     ' wind due to azimuths used assuming 1 m/s'
        #                     ' error in radial velocities')
        #u_error.units = 'm/s'
        #v_error = nc_file.createVariable('v_error', 'f', ('time', 'height'))
        #v_error.missing_value = -9999.0
        #v_error[:, :] = self.dv
        #v_error.long_name = ('Sampling uncertainty in northward component of'
        #                     ' wind due to azimuths used assuming 1 m/s'
        #                     ' error in radial velocities')
        #v_error.units = 'm/s'
        #w_error = nc_file.createVariable('w_error', 'f', ('time', 'height'))
        #w_error.missing_value = -9999.0
        #w_error[:, :] = self.dw
        #w_error.long_name = ('Sampling uncertainty in vertical component of'
        #                     ' wind due to azimuths used assuming 1 m/s'
        #                     ' error in radial velocities')
        #w_error.units = 'm/s'
        # wspd/wdir errors (not currently used)
        # wind_speed_error = nc_file.createVariable('wind_speed_error', 'f',
        #                                           ('time', 'height'))
        # wind_speed_error.missing_value = -9999.0
        # # not currently calculating wind speed error?
        # wind_speed_error[:, :] = wind_speed[:, :] * np.nan
        # wind_speed_error.long_name = 'Wind speed error'
        # wind_speed_error.units = 'm/s'
        # wind_direction_error = nc_file.createVariable('wind_direction_error',
        #                                               'f', ('time', 'height'))
        # wind_direction_error.missing_value = -9999.0
        # wind_direction_error[:, :] = wind_direction[:, :] * np.nan
        # wind_direction_error.long_name = 'Wind direction error'
        # wind_direction_error.units = 'm/s'
        # scan metadata
        scan_duration = nc_file.createVariable('scan_duration', 'f', 'time')
        scan_duration.missing_value = -9999.0
        scan_duration[:] = [(i[0] - i[1]).total_seconds()
                            for i in zip(self.etime, self.stime)]
        scan_duration.long_name = 'PPI scan duration'
        scan_duration.units = 'second'
        elevation_angle = nc_file.createVariable('elevation_angle', 'f',
                                                 'time')
        elevation_angle.missing_value = -9999.0
        elevation_angle[:] = self.el
        elevation_angle.long_name = 'Beam elevation angle'
        elevation_angle.units = 'degree'
        nbeams = nc_file.createVariable('nbeams', 'i', 'time')
        nbeams[:] = self.nbeams
        nbeams.long_name = ('Number of beams (azimuth angles) in each PPI')
        nbeams.units = 'unitless'
        nbeams_used = nc_file.createVariable('nbeams_used', 'i', ('time',
                                                                  'height'))
        nbeams_used.missing_value = -9999.0
        nbeams_used[:, :] = self.nbeams_used
        nbeams_used.long_name = ('Number of beams (azimuth angles) used in '
                                 'wind vector estimations')
        nbeams_used.units = 'unitless'

    def add_attributes(self, nc_file: netCDF4.Dataset):
        """
        Add metadata attributes to netCDF
        """
        nc_file.input_ppi_min_cnr = self.min_cnr
        # add thresholding vals as attributes, if present
        if self.thresholds:
            for t in self.thresholds:
                name = "threshold_" + t
                nc_file.setncattr(name, self.thresholds[t])



import numpy as np
import numpy.ma as ma
import datetime as dt
import pytz
from typing import List
import netCDF4

import iss_lidar.tools as tools
from .vad import VADSet
from .Lidar_functions import consensus_avg


class ConsensusSet(VADSet):

    @staticmethod
    def create_time_ranges(day: dt.date, size: dt.timedelta
                           ) -> List[dt.datetime]:
        """ Create a list of datetimes every given interval for this day """
        # currently if the interval size is not divisible into 24 hours, the
        # last time range will extend into the following day
        start = dt.datetime(day.year, day.month, day.day, tzinfo=pytz.UTC)
        end = start + dt.timedelta(days=1)
        ranges = []
        while (start < end):
            ranges.append(start)
            start = start + size
        return ranges

    @staticmethod
    def median_from_consensus_idxs(vals: ma.array, idxs: List) -> float:
        """
        Calculate the median of an array from the subset of elements at the
        same indices as were used for consensus averaging for w.
        """
        # avoid "mean of an empty slice" numpy runtime warning
        if not idxs:
            return np.nan
        return ma.median(vals[idxs])

    def __init__(self, alt: float, lat: float, lon: float, height: np.array,
                 time: list, u: ma.array, v: ma.array, w: ma.array,
                 n_u: ma.array, n_v: ma.array, n_w: ma.array,
                 residual: ma.array, correlation: ma.array, mean_cnr: ma.array,
                 wspd: ma.array, wdir: ma.array, window: float,
                 span: dt.timedelta):
        # separating out calculating from vadset into a classmethod in case we
        # want another classmethod to read from file in the future
        self.alt = alt
        self.lat = lat
        self.lon = lon
        self.height = height
        self.stime = time
        # generate etime from stime and span, no need to supply it
        self.etime = [s + span for s in self.stime]
        self.u = u
        self.v = v
        self.w = w
        self.n_u = n_u  # number of indices averaged in consensus
        self.n_v = n_v
        self.n_w = n_w
        self.residual = residual
        self.correlation = correlation
        self.mean_cnr = mean_cnr
        self.speed = wspd
        self.wdir = wdir
        self.window = window  # size of consensus buckets
        self.span = span  # timespan to average over

    @classmethod
    def from_VADSet(cls, vs: VADSet, window: float, span: dt.timedelta):
        # create half-hour time bins
        ranges = ConsensusSet.create_time_ranges(vs.stime[0].date(), span)
        u_mean = ma.zeros((len(ranges), len(vs.height)))
        v_mean = ma.zeros((len(ranges), len(vs.height)))
        w_mean = ma.zeros((len(ranges), len(vs.height)))
        n_u = ma.zeros((len(ranges), len(vs.height)))
        n_v = ma.zeros((len(ranges), len(vs.height)))
        n_w = ma.zeros((len(ranges), len(vs.height)))
        residual = ma.zeros((len(ranges), len(vs.height)))
        correlation = ma.zeros((len(ranges), len(vs.height)))
        mean_cnr = ma.zeros((len(ranges), len(vs.height)))

        for idx, r in enumerate(ranges):
            start = r
            end = start + span
            bin_idxs = [i for i in range(len(vs.stime)) if vs.stime[i] >= start
                        and vs.stime[i] < end]

            # no vads in this time window
            if len(bin_idxs) == 0:
                u_mean[idx, :] = np.nan
                v_mean[idx, :] = np.nan
                w_mean[idx, :] = np.nan
                residual[idx, :] = np.nan
                correlation[idx, :] = np.nan
                mean_cnr[idx, :] = np.nan
                n_u[idx, :] = np.nan
                n_v[idx, :] = np.nan
                n_w[idx, :] = np.nan
                continue

            u_bin = ma.array([vs.u[i] for i in bin_idxs])
            v_bin = ma.array([vs.v[i] for i in bin_idxs])
            w_bin = ma.array([vs.w[i] for i in bin_idxs])
            res_bin = ma.array([vs.residual[i] for i in bin_idxs])
            cor_bin = ma.array([vs.correlation[i] for i in bin_idxs])
            mc_bin = ma.array([vs.mean_cnr[i] for i in bin_idxs])

            for hgt in range(len(vs.height)):
                avg, idxs = consensus_avg(u_bin[:, hgt], window)
                u_mean[idx, hgt] = avg
                n_u[idx, hgt] = len(idxs)
                avg, idxs = consensus_avg(v_bin[:, hgt], window)
                v_mean[idx, hgt] = avg
                n_v[idx, hgt] = len(idxs)
                avg, idxs = consensus_avg(w_bin[:, hgt], window)
                w_mean[idx, hgt] = avg
                n_w[idx, hgt] = len(idxs)
                # take median of idxs used for w to avg res, corr, mean_cnr
                residual[idx, hgt] = ConsensusSet.\
                    median_from_consensus_idxs(res_bin[:, hgt], idxs)
                correlation[idx, hgt] = ConsensusSet.\
                    median_from_consensus_idxs(cor_bin[:, hgt], idxs)
                mean_cnr[idx, hgt] = ConsensusSet.\
                    median_from_consensus_idxs(mc_bin[:, hgt], idxs)

        # calculate wspd/wdir from averaged u and v
        wspd, wdir = tools.wspd_wdir_from_uv(u_mean, v_mean)

        return cls(vs.alt,
                   vs.lat,
                   vs.lon,
                   vs.height,
                   ranges,
                   ma.masked_invalid(u_mean),
                   ma.masked_invalid(v_mean),
                   ma.masked_invalid(w_mean),
                   ma.masked_invalid(n_u),
                   ma.masked_invalid(n_v),
                   ma.masked_invalid(n_w),
                   ma.masked_invalid(residual),
                   ma.masked_invalid(correlation),
                   ma.masked_invalid(mean_cnr),
                   ma.masked_invalid(wspd),
                   ma.masked_invalid(wdir),
                   window,
                   span)

    @classmethod
    def from_file(cls, filename: str):
        """ Create a ConsensusSet object from a daily consensus netcdf file """
        f = netCDF4.Dataset(filename, 'r')
        stime = list(netCDF4.num2date(f.variables['time'][:],
                                      f.variables['time'].units,
                                      only_use_python_datetimes=True,
                                      only_use_cftime_datetimes=False))
        # reconstitute endtime from scan duration
        # make dates tz aware
        stime = [s.replace(tzinfo=pytz.utc) for s in stime]
        # scalar values with a missing_value get read in as np.ma.MaskedArray
        # if the single value is not masked, but as np.ma.core.MaskedConstant
        # if the single value is masked. Instances of MaskedConstant don't
        # compare well to each other or to other data types, so if the scalar
        # lat/lon/alt is masked, replace it with a null-dimensioned masked
        # array with a value of nan.
        alt = (ma.array(np.nan) if f.variables['alt'][:] is np.ma.masked
               else f.variables['alt'][:])
        lat = (ma.array(np.nan) if f.variables['lat'][:] is np.ma.masked
               else f.variables['lat'][:])
        lon = (ma.array(np.nan) if f.variables['lon'][:] is np.ma.masked
               else f.variables['lon'][:])
        # reconstitute span by finding the difference between subsequent times
        span = stime[1] - stime[0]
        return cls(alt,
                   lat,
                   lon,
                   f.variables['height'][:],
                   stime,
                   ma.array(f.variables['u'][:]),
                   ma.array(f.variables['v'][:]),
                   ma.array(f.variables['w'][:]),
                   ma.array(f.variables['u_npoints'][:]),
                   ma.array(f.variables['v_npoints'][:]),
                   ma.array(f.variables['w_npoints'][:]),
                   ma.array(f.variables['residual'][:]),
                   ma.array(f.variables['correlation'][:]),
                   ma.array(f.variables['mean_snr'][:]),
                   ma.array(f.variables['wind_speed'][:]),
                   ma.array(f.variables['wind_direction'][:]),
                   int(f.consensus_avg_window),
                   span)

    def add_aux_variables(self, nc_file: netCDF4.Dataset):
        # number of points used in consensus for uvw
        n_u = nc_file.createVariable('u_npoints', 'f', ('time', 'height'))
        n_u.missing_value = -9999.0
        n_u[:, :] = self.n_u
        n_u.long_name = ('Number of points used in consensus averaging window '
                         'for eastward component of winds')
        n_u.units = 'unitless'
        n_v = nc_file.createVariable('v_npoints', 'f', ('time', 'height'))
        n_v.missing_value = -9999.0
        n_v[:, :] = self.n_v
        n_v.long_name = ('Number of points used in consensus averaging window '
                         'for northward component of winds')
        n_v.units = 'unitless'
        n_w = nc_file.createVariable('w_npoints', 'f', ('time', 'height'))
        n_w.missing_value = -9999.0
        n_w[:, :] = self.n_w
        n_w.long_name = ('Number of points used in consensus averaging window '
                         'for vertical component of winds')
        n_w.units = 'unitless'

    def add_attributes(self, nc_file: netCDF4.Dataset):
        nc_file.consensus_avg_window = self.window


# Object to hold data from a PPI scan

import numpy as np
import numpy.ma as ma
from datetime import datetime
from .tools import read_cfradial


class PPI:

    def __init__(self, cnr: np.ma.MaskedArray = None,
                 ranges: np.ma.MaskedArray = None,
                 vr: np.ma.MaskedArray = None,
                 elevation: np.ma.MaskedArray = None,
                 azimuth: np.ma.MaskedArray = None, starttime: datetime = None,
                 endtime: datetime = None,
                 lat: np.ma.core.MaskedConstant = None,
                 lon: np.ma.core.MaskedConstant = None,
                 alt: np.ma.core.MaskedConstant = None):
        """ Create PPI object from numpy arrays of data"""
        self.cnr = cnr
        self.ranges = ranges
        self.vr = vr
        self.elevation = elevation
        self.azimuth = azimuth
        self.starttime = starttime
        self.endtime = endtime
        self.lat = lat
        self.lon = lon
        self.alt = alt

    @classmethod
    def from_file(cls, filename: str):
        """ Create PPI object by reading cfradial scan file """
        [cnr, ranges, vr, elevation, azimuth, starttime, endtime, lat, lon,
         alt] = read_cfradial(filename)
        return cls(cnr, ranges, vr, elevation, azimuth, starttime, endtime,
                   lat, lon, alt)

    def threshold_cnr(self, max_cnr: int):
        """ Set vr to nan if cnr is below threshold """
        self.vr = ma.masked_where(self.cnr < max_cnr, self.vr)

    def mean_cnr(self) -> np.ndarray:
        """ Return mean value of CNR """
        return np.nanmean(self.cnr, axis=0)



import numpy.ma as ma
from iss_lidar.ppi import PPI


def test_threshold_cnr():
    vel = ma.arange(1, 20)
    ppi = PPI(cnr=ma.arange(-29, -10), vr=vel)
    ppi.threshold_cnr(-22)
    vel = ma.masked_array(vel, [1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0,
                                0, 0, 0, 0])
    assert ma.allequal(vel, ppi.vr)
